# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging
from concurrent.futures import Future
from functools import partial
from pathlib import Path

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common.enum import AudioPlayerState
from gajim.common.multiprocess.audio_preview import extract_audio_properties
from gajim.common.util.text import format_duration

from gajim.gtk.audio_player import AudioPlayer
from gajim.gtk.preview.audio_visualizer import AudioVisualizerWidget
from gajim.gtk.preview.misc import LoadingBox  # noqa: F401 # pyright: ignore
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string

log = logging.getLogger("gajim.gtk.preview.audio")


SEEK_BAR_PADDING = 12


@Gtk.Template.from_string(string=get_ui_string("preview/audio.ui"))
class AudioPreviewWidget(Gtk.Box, SignalManager):
    __gtype_name__ = "AudioPreviewWidget"

    __gsignals__ = {
        "display-error": (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            None,
            (),
        )
    }

    _open_folder_button: Gtk.Button = Gtk.Template.Child()
    _save_as_button: Gtk.Button = Gtk.Template.Child()
    _stack: Gtk.Stack = Gtk.Template.Child()
    _seek_bar_adj: Gtk.Adjustment = Gtk.Template.Child()
    _speed_bar_adj: Gtk.Adjustment = Gtk.Template.Child()
    _drawing_box: Gtk.Box = Gtk.Template.Child()
    _seek_bar_box: Gtk.Box = Gtk.Template.Child()
    _seek_bar: Gtk.Scale = Gtk.Template.Child()
    _progress_label: Gtk.Label = Gtk.Template.Child()
    _control_box: Gtk.Box = Gtk.Template.Child()
    _rewind_button: Gtk.Button = Gtk.Template.Child()
    _play_pause_button: Gtk.Button = Gtk.Template.Child()
    _play_icon: Gtk.Image = Gtk.Template.Child()
    _forward_button: Gtk.Button = Gtk.Template.Child()
    _speed_dec_button: Gtk.Button = Gtk.Template.Child()
    _speed_menubutton: Gtk.MenuButton = Gtk.Template.Child()
    _speed_label: Gtk.Label = Gtk.Template.Child()
    _speed_inc_button: Gtk.Button = Gtk.Template.Child()
    _speed_popover: Gtk.Popover = Gtk.Template.Child()
    _speed_bar: Gtk.Scale = Gtk.Template.Child()

    def __init__(
        self,
        audio_player: AudioPlayer,
        filename: str,
        file_size: int,
        orig_path: Path,
    ) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)

        self._audio_player = audio_player
        self._orig_path = orig_path
        self._filename = filename
        self._file_size = file_size

        self._id = hash(self._orig_path) % 100_000

        self._preview_state = self._audio_player.get_audio_state(self._id)

        # UI
        self._is_ltr = bool(self.get_direction() == Gtk.TextDirection.LTR)
        self._offset_backward = -10e9  # in ns
        self._offset_forward = 10e9
        self._cursor_pos = 0.0
        self._user_holds_position_slider = False
        self._seek_ts = -1  # in ns, -1 means invalid
        self._new_voice_message_track = False

        self._open_folder_button.set_action_target_value(
            GLib.Variant("s", str(self._orig_path))
        )
        self._save_as_button.set_action_target_value(
            GLib.Variant("s", str(self._orig_path))
        )

        self._speed_bar_adj.set_value(self._preview_state.speed)
        self._set_playback_speed(self._preview_state.speed)

        self._speed_bar.add_mark(0.25, Gtk.PositionType.BOTTOM, "0.25")
        self._speed_bar.add_mark(0.5, Gtk.PositionType.BOTTOM, "")
        self._speed_bar.add_mark(0.75, Gtk.PositionType.BOTTOM, "")
        self._speed_bar.add_mark(1, Gtk.PositionType.BOTTOM, "1.0")
        self._speed_bar.add_mark(1.25, Gtk.PositionType.BOTTOM, "")
        self._speed_bar.add_mark(1.5, Gtk.PositionType.BOTTOM, "1.5")
        self._speed_bar.add_mark(1.75, Gtk.PositionType.BOTTOM, "")
        self._speed_bar.add_mark(2, Gtk.PositionType.BOTTOM, "2")
        self._connect(self._speed_bar, "change-value", self._on_speed_change)

        self._connect(self._speed_dec_button, "clicked", self._on_speed_dec_clicked)
        self._connect(self._speed_inc_button, "clicked", self._on_speed_inc_clicked)

        self._connect(self._rewind_button, "clicked", self._on_rewind_clicked)
        self._connect(self._forward_button, "clicked", self._on_forward_clicked)

        self._gesture_visualizer_click = Gtk.GestureClick(
            button=Gdk.BUTTON_PRIMARY, propagation_phase=Gtk.PropagationPhase.CAPTURE
        )
        self._connect(
            self._gesture_visualizer_click,
            "pressed",
            self._on_visualizer_button_clicked,
        )

        gesture_seek_click = Gtk.GestureClick(
            button=Gdk.BUTTON_PRIMARY, propagation_phase=Gtk.PropagationPhase.CAPTURE
        )
        self._connect(gesture_seek_click, "pressed", self._on_seek_bar_button_pressed)
        self._connect(gesture_seek_click, "released", self._on_seek_bar_button_released)
        self._seek_bar_box.add_controller(gesture_seek_click)

        controller_motion = Gtk.EventControllerMotion(
            propagation_phase=Gtk.PropagationPhase.CAPTURE
        )
        self._connect(controller_motion, "motion", self._on_seek_bar_cursor_move)
        self._seek_bar.add_controller(controller_motion)

        controller_scroll = Gtk.EventControllerScroll(
            flags=Gtk.EventControllerScrollFlags.VERTICAL
        )
        self._connect(controller_scroll, "scroll", self._on_seek_bar_scrolled)
        self._seek_bar.add_controller(controller_scroll)

        gesture_timestamp_click = Gtk.GestureClick(
            button=Gdk.BUTTON_PRIMARY, propagation_phase=Gtk.PropagationPhase.TARGET
        )
        self._connect(
            gesture_timestamp_click, "pressed", self._on_timestamp_label_clicked
        )
        self._progress_label.add_controller(gesture_timestamp_click)

        self._visualizer = AudioVisualizerWidget(
            width=300 - 2 * SEEK_BAR_PADDING, x_offset=SEEK_BAR_PADDING
        )
        self._visualizer.add_controller(self._gesture_visualizer_click)
        self._drawing_box.append(self._visualizer)

        self._progress_id = None
        self._connect(self._play_pause_button, "clicked", self._on_play_clicked)

        if self._orig_path.is_file():
            # In case of the voice message recorder, start with no existing file
            if self._preview_state.duration == 0 or not self._preview_state.samples:
                self._get_audio_properties()
            else:
                self._display_audio_preview()

    @property
    def id(self) -> int:
        return self._id

    def do_unroot(self) -> None:
        Gtk.Box.do_unroot(self)
        self._disconnect_all()
        self._audio_player.stop(self._id)
        del self._audio_player

    def sample_voice_message(self, audio_path: Path) -> None:
        self._orig_path = audio_path
        self._preview_state.position = 0
        self._seek_bar.set_value(0)
        self._new_voice_message_track = True
        self._get_audio_properties()

    def _get_audio_properties(self) -> None:
        assert self._orig_path is not None

        try:
            future = app.process_pool.submit(extract_audio_properties, self._orig_path)
            future.add_done_callback(
                partial(GLib.idle_add, self._get_audio_properties_finished)
            )
        except Exception as error:
            log.warning("Creating thumbnail failed for: %s %s", self._orig_path, error)

    def _get_audio_properties_finished(
        self, future: Future[tuple[list[tuple[float, float]], int]]
    ) -> bool:
        try:
            samples, duration = future.result()
        except Exception as error:
            log.exception(
                "Creating thumbnail failed for: %s %s", self._orig_path, error
            )
        else:
            self._preview_state.samples = samples
            self._preview_state.duration = duration
            log.info("State: %s", self._preview_state)
            self._display_audio_preview()

        return GLib.SOURCE_REMOVE

    def _set_preview_state_position(self, position: float) -> None:
        position = max(0.0, position)
        position = min(position, self._preview_state.duration)
        self._preview_state.position = position

    def _convert_pos_to_text(self, current_position: float, duration: float):
        dur_str = format_duration(duration, duration)
        ltr_char = "\u200e"

        if self._preview_state.is_timestamp_positive:
            cur_str = f"{format_duration(current_position, duration)}"
            cur_str = f"{cur_str}/{dur_str}"
        else:
            cur_str = f"{format_duration(duration - current_position, duration)}"
            cur_str = f"{ltr_char}-{cur_str}/{dur_str}"

        return cur_str

    def _on_audio_playback_progressed(
        self, _obj: AudioPlayer, preview_id: int, _state: AudioPlayerState
    ) -> None:
        if preview_id != self.id:
            return
        self._update_ui_from_state()

    def _on_audio_playback_changed(
        self, _obj: AudioPlayer, preview_id: int, state: int
    ) -> None:
        if preview_id != self.id:
            self._disconnect_object(self._audio_player)
            self._play_icon.set_from_icon_name("lucide-play-symbolic")
            return

        if state == AudioPlayerState.PLAYING.value:
            self._play_icon.set_from_icon_name("lucide-pause-symbolic")
        else:
            self._play_icon.set_from_icon_name("lucide-play-symbolic")

        self._update_ui_from_state()

    def _update_ui_from_state(self):
        if not self._user_holds_position_slider:
            self._seek_bar.set_value(self._preview_state.position)
        self._visualizer.render_static_graph(
            self._preview_state.position / self._preview_state.duration,
            self._seek_ts / self._preview_state.duration,
        )
        self._progress_label.set_text(
            self._convert_pos_to_text(
                self._preview_state.position, self._preview_state.duration
            )
        )

    def _on_play_clicked(self, _button: Gtk.Button) -> None:
        if self._audio_player.preview_id != self._id or self._new_voice_message_track:
            self._new_voice_message_track = False
            self._connect(
                self._audio_player,
                "audio-playback-progressed",
                self._on_audio_playback_progressed,
            )
            self._connect(
                self._audio_player,
                "audio-playback-changed",
                self._on_audio_playback_changed,
            )
            self._audio_player.play_audio_file(self._orig_path, self._id)
        else:
            self._audio_player.toggle_playback()

    def _update_timestamp_label(self) -> None:
        cur = self._preview_state.position
        dur = self._preview_state.duration

        dur_str = format_duration(dur, dur)
        ltr_char = "\u200e"

        if self._preview_state.is_timestamp_positive:
            cur_str = f"{format_duration(cur, dur)}"
            self._progress_label.set_text(f"{cur_str}/{dur_str}")
        else:
            cur_str = f"{format_duration(dur - cur, dur)}"
            self._progress_label.set_text(f"{ltr_char}-{cur_str}/{dur_str}")

    def _on_timestamp_label_clicked(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        _x: float,
        _y: float,
    ) -> None:
        gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)
        self._preview_state.is_timestamp_positive = (
            not self._preview_state.is_timestamp_positive
        )
        self._update_timestamp_label()

    def _convert_position_to_timestamp(self, x: float, x_max: float) -> float:
        if x_max == 0.0:
            log.warning("Width is zero, when converting position to timestamp")
            return 0.0

        if not self._is_ltr:
            x = x_max - x
        x = max(0.0, x)
        x = min(x, x_max)
        timestamp = x / x_max * self._preview_state.duration
        return timestamp

    def _on_seek_bar_button_pressed(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        _x: float,
        _y: float,
    ) -> None:
        self._user_holds_position_slider = True

    def _on_seek_bar_button_released(
        self,
        _gesture_click: Gtk.GestureClick,
        _n_press: int,
        _x: float,
        _y: float,
    ) -> None:
        seek_ts = self._convert_position_to_timestamp(
            self._cursor_pos, self._seek_bar.get_width()
        )
        if self._audio_player.preview_id != self.id:
            self._preview_state.position = seek_ts
        else:
            self._audio_player.set_playback_position(self._id, seek_ts)

        self._update_ui_from_state()

        self._seek_ts = -1
        self._user_holds_position_slider = False

    def _on_seek_bar_cursor_move(
        self,
        _event_controller: Gtk.EventControllerMotion,
        x: float,
        _y: float,
    ) -> None:
        if not self._user_holds_position_slider:
            return

        if abs(x - self._cursor_pos) < 1e-2:
            return

        self._cursor_pos = x
        seek_ts = self._convert_position_to_timestamp(x, self._seek_bar.get_width())
        if (
            self._user_holds_position_slider
            and self._preview_state.pipeline_state != AudioPlayerState.PLAYING
        ):
            if self._audio_player.preview_id != self.id:
                self._preview_state.position = seek_ts
            else:
                self._audio_player.set_playback_position(self._id, seek_ts)
            self._update_ui_from_state()
        else:
            self._seek_ts = seek_ts

    def _on_seek_bar_scrolled(
        self,
        _event_controller: Gtk.EventControllerScroll,
        _dx: float,
        dy: float,
    ) -> None:
        if dy > 0:
            timestamp = self._preview_state.position + self._offset_backward
        else:
            timestamp = self._preview_state.position + self._offset_forward

        self._set_preview_state_position(timestamp)
        self._audio_player.set_playback_position(self.id, self._preview_state.position)
        self._update_ui_from_state()

    def _on_visualizer_button_clicked(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        _y: float,
    ) -> None:
        assert self._cursor_pos is not None

        gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)
        timestamp = self._convert_position_to_timestamp(
            x, self._visualizer.get_effective_width()
        )

        self._set_preview_state_position(timestamp)
        self._update_ui_from_state()
        self._audio_player.set_playback_position(self._id, timestamp)

    def _display_audio_preview(self) -> None:
        self._visualizer.set_samples(self._preview_state.samples)
        self._seek_bar_adj.set_lower(0.0)
        self._seek_bar_adj.set_upper(self._preview_state.duration)
        self._update_ui_from_state()
        self._stack.set_visible_child_name("preview")

    def _set_playback_speed(self, speed: float) -> None:
        speed = max(0.25, speed)
        speed = min(speed, 2.0)

        self._preview_state.speed = speed
        self._speed_label.set_text(f"{speed:.2f}x")
        self._speed_bar_adj.set_value(speed)

        if self._audio_player.preview_id != self.id:
            return

        self._audio_player.set_playback_speed(speed)

    def _on_speed_change(
        self, _range: Gtk.Range, _scroll: Gtk.ScrollType, value: float
    ) -> None:
        self._set_playback_speed(value)

    def _on_speed_dec_clicked(self, _button: Gtk.Button) -> None:
        speed = self._preview_state.speed - self._speed_bar_adj.get_step_increment()
        self._set_playback_speed(speed)

    def _on_speed_inc_clicked(self, _button: Gtk.Button) -> None:
        speed = self._preview_state.speed + self._speed_bar_adj.get_step_increment()
        self._set_playback_speed(speed)

    def _change_playback_position_by_step(self, step: float) -> None:
        new_pos = self._preview_state.position + step
        new_pos = max(0.0, new_pos)
        new_pos = min(new_pos, self._preview_state.duration)

        if self._audio_player.preview_id != self.id:
            self._preview_state.position = new_pos
        else:
            self._audio_player.set_playback_position(self._id, new_pos)

        self._update_ui_from_state()

    def _on_rewind_clicked(self, _button: Gtk.Button) -> None:
        self._change_playback_position_by_step(self._offset_backward)

    def _on_forward_clicked(self, _button: Gtk.Button) -> None:
        self._change_playback_position_by_step(self._offset_forward)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import time
from datetime import datetime

from gi.repository import Gtk
from gi.repository import Pango
from nbxmpp.structs import MucSubject

from gajim.common.const import AvatarSize
from gajim.common.i18n import _
from gajim.common.util.text import process_non_spacing_marks

from gajim.gtk.conversation.message_widget import MessageWidget
from gajim.gtk.conversation.rows.base import BaseRow
from gajim.gtk.conversation.rows.widgets import DateTimeLabel


class MUCSubject(BaseRow):
    type = "muc-subject"

    def __init__(
        self, account: str, subject: MucSubject, timestamp: float | None = None
    ) -> None:
        BaseRow.__init__(self, account)

        current_timestamp = timestamp or time.time()
        self.timestamp = datetime.fromtimestamp(current_timestamp).astimezone()
        self.db_timestamp = current_timestamp

        self.grid.set_halign(Gtk.Align.START)

        avatar_placeholder = Gtk.Box()
        avatar_placeholder.set_size_request(AvatarSize.ROSTER, -1)
        self.grid.attach(avatar_placeholder, 0, 0, 1, 2)

        subject_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        subject_box.set_halign(Gtk.Align.START)
        subject_box.add_css_class("card")
        subject_box.add_css_class("p-12")
        subject_box.add_css_class("conversation-subject-box")

        title = Gtk.Label(label=_("Subject"))
        title.set_halign(Gtk.Align.START)
        title.add_css_class("bold")
        subject_box.append(title)

        author = _("Changed by %s") % (subject.author or _("Unknown"))

        date = ""
        if subject.timestamp is not None:
            time_str = time.strftime("%c", time.localtime(subject.timestamp))
            date = f"{time_str}\n"

        meta_str = f"{author}\n{date}"
        meta = Gtk.Label(
            halign=Gtk.Align.START,
            label=process_non_spacing_marks(meta_str),
            selectable=True,
            wrap=True,
            wrap_mode=Pango.WrapMode.WORD_CHAR,
            xalign=0,
        )
        subject_box.append(meta)

        message_widget = MessageWidget(account)
        message_widget.add_with_styling(subject.text)
        subject_box.append(message_widget)
        self.grid.attach(subject_box, 1, 0, 1, 1)

        timestamp_widget = DateTimeLabel(self.timestamp)
        timestamp_widget.set_valign(Gtk.Align.START)
        self.grid.attach(timestamp_widget, 2, 0, 1, 1)

    def do_unroot(self) -> None:
        BaseRow.do_unroot(self)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import GObject
from gi.repository import Gtk

from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant

ContactT = BareContact | GroupchatContact | GroupchatParticipant


class JumpToEndButton(Gtk.Overlay):
    __gsignals__ = {
        "clicked": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self) -> None:
        Gtk.Overlay.__init__(self)
        self.set_halign(Gtk.Align.END)
        self.set_valign(Gtk.Align.END)
        self.set_margin_end(6)
        self.set_margin_bottom(12)

        icon = Gtk.Image.new_from_icon_name("lucide-chevrons-down-symbolic")
        icon.set_margin_start(4)
        icon.set_margin_end(4)

        button = Gtk.Button()
        button.set_margin_top(6)
        button.set_margin_end(6)
        button.add_css_class("circular")
        button.set_child(icon)
        button.connect("clicked", self._on_jump_clicked)
        self.set_child(button)

        self._unread_label = Gtk.Label()
        self._unread_label.add_css_class("unread-counter")
        self._unread_label.set_visible(False)
        self._unread_label.set_halign(Gtk.Align.END)
        self._unread_label.set_valign(Gtk.Align.START)
        self.add_overlay(self._unread_label)

        self._count = 0
        self.set_visible(False)

    def switch_contact(self, contact: ContactT) -> None:
        if isinstance(contact, GroupchatContact) and not contact.can_notify():
            self._unread_label.add_css_class("unread-counter-silent")
        else:
            self._unread_label.remove_css_class("unread-counter-silent")

    def _on_jump_clicked(self, _button: Gtk.Button) -> None:
        self.reset_unread_count()
        self.emit("clicked")

    def toggle(self, visible: bool) -> None:
        self.set_visible(visible)
        if not visible:
            self.reset_unread_count()

    def reset_unread_count(self) -> None:
        self._count = 0
        self._unread_label.set_visible(False)

    def add_unread_count(self) -> None:
        self._count += 1
        if self._count > 0:
            self._unread_label.set_text(str(self._count))
            self._unread_label.set_visible(True)
        else:
            self._unread_label.set_visible(False)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Literal
from typing import overload

from collections.abc import Callable
from dataclasses import dataclass

from gi.repository import Adw
from gi.repository import Gtk

from gajim.common import app
from gajim.common.i18n import _

from gajim.gtk.util.classes import SignalManager

AppearanceT = Literal["default", "suggested", "destructive"]


@dataclass
class DialogResponse:
    response_id: str
    label: str
    appearance: AppearanceT = "default"
    is_default: bool = False


@dataclass
class CancelDialogResponse(DialogResponse):
    response_id: str = "cancel"
    label: str = _("_Cancel")
    appearance: AppearanceT = "default"
    is_default: bool = True


class DialogEntry(Gtk.Entry):
    def __init__(self, text: str = ""):
        Gtk.Entry.__init__(self, margin_start=50, margin_end=50)
        self.set_text(text)

    def get_value(self) -> str:
        return self.get_text()


class DialogCheckButton(Gtk.Box):
    def __init__(self, label: str):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)

        label_widget = Gtk.Label(
            wrap=True,
            max_width_chars=50,
            halign=Gtk.Align.START,
            margin_start=10,
        )
        label_widget.set_text_with_mnemonic(label)

        self._checkbutton = Gtk.CheckButton(
            child=label_widget,
            can_focus=False,
            margin_start=30,
            margin_end=30,
        )

        self.append(self._checkbutton)

    def get_value(self) -> bool:
        return self._checkbutton.get_active()


ExtraWidgetT = DialogEntry | DialogCheckButton | None


class _BaseAlertDialog(Adw.AlertDialog, SignalManager):
    def __init__(
        self,
        heading: str,
        body: str,
        *,
        body_use_markup: bool = False,
        responses: list[DialogResponse] | None = None,
        close_response: str = "close",
        emit_responses: list[str] | None = None,
        extra_widget: ExtraWidgetT = None,
        callback: Callable[..., None] | None = None,
        parent: Gtk.Window | None = None,
    ) -> None:
        SignalManager.__init__(self)

        if parent is None:
            parent = app.app.get_active_window()

        if responses is None:
            responses = []

        self._callback = callback
        self._emit_responses = emit_responses

        Adw.AlertDialog.__init__(
            self,
            heading=heading,
            body=body,
            body_use_markup=body_use_markup,
            prefer_wide_layout=True,
            close_response=close_response,
        )

        for response in responses:
            self.add_response(response.response_id, response.label)
            match response.appearance:
                case "suggested":
                    self.set_response_appearance(
                        response.response_id, Adw.ResponseAppearance.SUGGESTED
                    )
                case "destructive":
                    self.set_response_appearance(
                        response.response_id, Adw.ResponseAppearance.DESTRUCTIVE
                    )
                case _:
                    pass

            if response.is_default:
                self.set_default_response(response.response_id)

        self.set_extra_child(extra_widget)

        self._connect(self, "response", self._on_response)
        self.present(parent)

        if extra_widget is not None:
            extra_widget.grab_focus()

    def _emit_response(self, response_id: str) -> None:
        if self._callback is None:
            return

        if self._emit_responses is not None and response_id not in self._emit_responses:
            return

        widget = cast(ExtraWidgetT, self.get_extra_child())

        args: list[Any] = []
        # Add the response_id as callback arg only if we emit more than one response
        if self._emit_responses is None or len(self._emit_responses) != 1:
            args.append(response_id)

        if widget is not None:
            args.append(widget.get_value())

        self._callback(*args)

    def _on_response(self, _dialog: Adw.AlertDialog, response_id: str) -> None:
        self._emit_response(response_id)
        self._cleanup()

    def _cleanup(self) -> None:
        self._disconnect_all()
        del self._callback


class AlertDialog(_BaseAlertDialog):
    pass


class InformationAlertDialog(_BaseAlertDialog):
    def __init__(
        self,
        heading: str,
        body: str,
        *,
        body_use_markup: bool = False,
        callback: Callable[[], None] | None = None,
        parent: Gtk.Window | None = None,
    ) -> None:
        _BaseAlertDialog.__init__(
            self,
            heading,
            body,
            body_use_markup=body_use_markup,
            responses=[
                DialogResponse("ok", _("_OK"), is_default=True),
            ],
            callback=callback,
            parent=parent,
        )

    def _emit_response(self, response_id: str) -> None:
        if self._callback is None:
            return

        self._callback()


class ConfirmationAlertDialog(_BaseAlertDialog):
    @overload
    def __init__(
        self,
        heading: str,
        body: str,
        confirm_label: str,
        *,
        body_use_markup: bool = ...,
        appearance: AppearanceT = ...,
        extra_widget: Literal[None] = ...,
        callback: Callable[[], None] | None = ...,
        parent: Gtk.Window | None = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        heading: str,
        body: str,
        confirm_label: str,
        *,
        body_use_markup: bool = ...,
        appearance: AppearanceT = ...,
        extra_widget: DialogEntry = ...,
        callback: Callable[[str], None] | None = ...,
        parent: Gtk.Window | None = ...,
    ) -> None: ...

    def __init__(
        self,
        heading: str,
        body: str,
        confirm_label: str,
        *,
        body_use_markup: bool = False,
        appearance: AppearanceT = "default",
        extra_widget: ExtraWidgetT = None,
        callback: Callable[..., None] | None = None,
        parent: Gtk.Window | None = None,
    ) -> None:
        _BaseAlertDialog.__init__(
            self,
            heading,
            body,
            body_use_markup=body_use_markup,
            responses=[
                CancelDialogResponse(),
                DialogResponse("confirm", confirm_label, appearance=appearance),
            ],
            emit_responses=["confirm"],
            extra_widget=extra_widget,
            callback=callback,
            parent=parent,
        )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import hashlib
import logging
import mimetypes
import os
from dataclasses import dataclass
from pathlib import Path
from urllib.parse import ParseResult
from urllib.parse import unquote
from urllib.parse import urlparse

from gi.repository import Gio
from gi.repository import GLib

from gajim.common import app
from gajim.common.const import ALL_MIME_TYPES
from gajim.common.const import AUDIO_MIME_TYPES
from gajim.common.const import IMAGE_MIME_TYPES
from gajim.common.const import VIDEO_MIME_TYPES
from gajim.common.helpers import sanitize_filename
from gajim.common.i18n import _
from gajim.common.i18n import p_
from gajim.common.regex import IRI_RX
from gajim.common.storage.archive import models as mod
from gajim.common.util.uri import Coords
from gajim.common.util.uri import get_geo_choords

log = logging.getLogger("gajim.c.util.preview")

MIME_TYPE_MAP = {"application/x-ext-webp": "image/webp"}


def get_image_paths(
    uri: str, urlparts: ParseResult, size: int, orig_dir: Path, thumb_dir: Path
) -> tuple[Path, Path]:
    path = Path(unquote(urlparts.path))
    web_stem = path.stem
    extension = path.suffix

    web_stem = sanitize_filename(web_stem)

    name_hash = hashlib.sha1(str(uri).encode()).hexdigest()

    orig_filename = f"{web_stem}_{name_hash}{extension}"

    thumb_filename = f"{web_stem}_{name_hash}_thumb_{size}.png"

    orig_path = orig_dir / orig_filename
    thumb_path = thumb_dir / thumb_filename
    return orig_path, thumb_path


def format_geo_coords(coords: Coords) -> str:
    lat = float(coords.lat)
    lon = float(coords.lon)

    def fmt(f: float) -> str:
        i = round(f * 3600.0)
        seconds = i % 60
        i //= 60
        minutes = i % 60
        i //= 60
        degrees = i
        return "%d°%02d′%02d′′" % (degrees, minutes, seconds)

    if lat >= 0:
        slat = p_("positive latitude", "%sN") % fmt(lat)
    else:
        slat = p_("negative latitude", "%sS") % fmt(-lat)
    if lon >= 0:
        slon = p_("positive longitude", "%sE") % fmt(lon)
    else:
        slon = p_("negative longitude", "%sW") % fmt(-lon)
    return f"{slat} {slon}"


def filename_from_uri(uri: str) -> str:
    urlparts = urlparse(unquote(uri))
    path = Path(urlparts.path)
    # Remove right-to-left override U+202E (commonly used to spoof extensions)
    return path.name.replace("\u202e", "")


def contains_audio_streams(file_path: Path) -> bool:
    # Check if it is really an audio file

    from gi.repository import GstPbutils

    has_audio = False
    discoverer = GstPbutils.Discoverer()
    try:
        info = discoverer.discover_uri(file_path.as_uri())
        has_audio = bool(info.get_audio_streams())
    except GLib.Error as err:
        log.error("Error while reading %s: %s", str(file_path), err)
        return False
    if not has_audio:
        log.warning("File does not contain audio stream: %s", str(file_path))
    return has_audio


def guess_mime_type(file_path: Path | str, data: bytes | None = None) -> str:
    file_path = str(file_path)

    # The mimetypes module maps extensions to mime types
    # it does no guessing based on file content
    mime_type, _ = mimetypes.guess_type(file_path)
    if mime_type is None:
        # Gio does also guess based on file content
        extension, _ = Gio.content_type_guess(file_path, data)
        mime_type = Gio.content_type_get_mime_type(extension)

    if mime_type is not None:
        # Mime type guessing via Gio on Windows is partly broken, which results in
        # e.g. "application/x-ext-webp". Try to map those mime types.
        # See https://gitlab.gnome.org/GNOME/librsvg/-/issues/1181#note_2526154
        mime_type = MIME_TYPE_MAP.get(mime_type, mime_type)

    log.debug("Guessed MIME type: %s", mime_type)
    return mime_type or ""


def guess_simple_file_type(
    file_path: str, data: bytes | None = None
) -> tuple[Gio.Icon, str, str]:
    mime_type = guess_mime_type(file_path, data)
    icon = get_icon_for_mime_type(mime_type)

    file_type = _("File")
    if mime_type.startswith("audio/"):
        file_type = _("Audio File")
    if mime_type.startswith("image/"):
        file_type = _("Image")
    if mime_type.startswith("video/"):
        file_type = _("Video")
    if mime_type.startswith("text/"):
        file_type = _("Text File")

    return icon, file_type, mime_type


def get_size_and_mime_type(path: Path) -> tuple[str, int]:
    return guess_mime_type(path), os.path.getsize(path)


def get_icon_for_mime_type(mime_type: str | None) -> Gio.Icon:
    if not mime_type:
        return Gio.Icon.new_for_string("lucide-file-symbolic")
    return Gio.content_type_get_icon(mime_type)


def is_audio(mime_type: str) -> bool:
    return mime_type in AUDIO_MIME_TYPES


def is_video(mime_type: str) -> bool:
    return mime_type in VIDEO_MIME_TYPES


def is_image(mime_type: str) -> bool:
    return mime_type in IMAGE_MIME_TYPES


@dataclass
class UrlPreview:
    uri: str
    mime_type: str
    file_name: str
    text: str
    icon: Gio.Icon
    encrypted: bool

    @classmethod
    def from_uri(cls, uri: str) -> UrlPreview:
        encrypted = uri.startswith("aesgcm://")
        file_name = filename_from_uri(uri)
        icon, file_type, mime_type = guess_simple_file_type(uri)
        text = f"{file_type} ({file_name})"
        return cls(
            uri=uri,
            mime_type=mime_type,
            file_name=file_name,
            text=text,
            icon=icon,
            encrypted=encrypted,
        )


@dataclass
class GeoPreview:
    uri: str
    text: str
    icon: Gio.Icon
    coords: Coords

    @classmethod
    def from_coords(cls, uri: str, coords: Coords) -> GeoPreview:
        icon = Gio.Icon.new_for_string("lucide-map-pin-symbolic")
        text = format_geo_coords(coords)
        return cls(uri=uri, text=text, icon=icon, coords=coords)


def get_preview_data(
    uri: str, oob_data: list[mod.OOB]
) -> GeoPreview | UrlPreview | None:
    if not IRI_RX.fullmatch(uri):
        # urlparse removes whitespace (and who knows what else) from URLs,
        # so can't be used for validation.
        return None

    try:
        urlparts = urlparse(uri)
    except Exception:
        return None

    if urlparts.scheme == "geo":
        coords = get_geo_choords(uri)
        if coords is not None:
            return GeoPreview.from_coords(uri, coords)
        return None

    if not urlparts.netloc:
        return None

    oob_url = None if not oob_data else oob_data[0].url

    if uri == oob_url or urlparts.scheme == "aesgcm":
        return UrlPreview.from_uri(uri)

    # http/https
    if urlparts.scheme not in ("https", "http"):
        log.info("Unsupported URI scheme: %s", uri)
        return None

    if not app.settings.get("preview_allow_all_images"):
        return None

    mime_type = guess_mime_type(uri)
    if mime_type not in ALL_MIME_TYPES:
        log.info("%s not in allowed mime types", mime_type)
        return None

    return UrlPreview.from_uri(uri)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from nbxmpp.protocol import InvalidJid
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID

from gajim.common import app


class InvalidFormat(Exception):
    pass


def parse_jid(jidstring: str) -> str:
    try:
        return str(validate_jid(jidstring))
    except Exception as error:
        raise InvalidFormat(error)


def parse_resource(resource: str) -> str | None:
    """
    Perform stringprep on resource and return it
    """
    if not resource:
        return None

    try:
        return resource.encode("OpaqueString").decode("utf-8")
    except UnicodeError:
        raise InvalidFormat("Invalid character in resource.")


def get_full_jid_from_iq(iq_obj: Iq) -> str | None:
    """
    Return the full jid (with resource) from an iq
    """
    jid = iq_obj.getFrom()
    if jid is None:
        return None
    return parse_jid(str(iq_obj.getFrom()))


def get_jid_from_iq(iq_obj: Iq) -> str | None:
    """
    Return the jid (without resource) from an iq
    """
    jid = get_full_jid_from_iq(iq_obj)
    if jid is None:
        return None
    return app.get_jid_without_resource(jid)


def validate_jid(jid: str | JID, type_: str | None = None) -> JID:
    try:
        jid = JID.from_string(str(jid))
    except InvalidJid as error:
        raise ValueError(error)

    if type_ is None:
        return jid
    if type_ == "bare" and jid.is_bare:
        return jid
    if type_ == "full" and jid.is_full:
        return jid
    if type_ == "domain" and jid.is_domain:
        return jid

    raise ValueError(f"Not a {type_} JID")



# Copyright (C) 2006 Gustavo Carneiro <gjcarneiro AT gmail.com>
#                    Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2006-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2008 Jean-Marie Traissard <jim AT lapin.org>
#                    Jonathan Schleifer <js-gajim AT webkeks.org>
#                    Stephan Erb <steve-e AT h3c.de>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import TypedDict

import logging
import sys

from gi.repository import Gio
from gi.repository import GLib
from nbxmpp.structs import TuneData

from gajim.common import app
from gajim.common.events import MusicTrackChanged

log = logging.getLogger("gajim.c.dbus.music_track")

MPRIS_PLAYER_PREFIX = "org.mpris.MediaPlayer2."


class TrackProperties(TypedDict, total=False):
    Metadata: dict[str, str]
    PlaybackStatus: str


class MusicTrackListener:
    _instance: MusicTrackListener | None = None

    @classmethod
    def get(cls) -> MusicTrackListener:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def __init__(self) -> None:
        self.players: dict[str, int] = {}
        self.connection: Gio.DBusConnection | None = None
        self._current_tune: TuneData | None = None
        self._running = False

        if sys.platform not in ("win32", "darwin"):
            self.start()

    def _emit(self, info: TuneData | None) -> None:
        self._current_tune = info
        app.ged.raise_event(MusicTrackChanged(info=info))

    @property
    def current_tune(self) -> TuneData | None:
        return self._current_tune

    def start(self):
        if self._running:
            return

        proxy = Gio.DBusProxy.new_for_bus_sync(
            Gio.BusType.SESSION,
            Gio.DBusProxyFlags.NONE,
            None,
            "org.freedesktop.DBus",
            "/org/freedesktop/DBus",
            "org.freedesktop.DBus",
            None,
        )

        self.connection = proxy.get_connection()
        self.connection.signal_subscribe(
            "org.freedesktop.DBus",
            "org.freedesktop.DBus",
            "NameOwnerChanged",
            "/org/freedesktop/DBus",
            None,
            Gio.DBusSignalFlags.NONE,
            self._signal_name_owner_changed,
        )

        try:
            result = proxy.call_sync(
                "ListNames", None, Gio.DBusCallFlags.NONE, -1, None
            )
        except GLib.Error as error:
            log.debug("Could not list names: %s", error.message)
            return

        for name in result[0]:
            if name.startswith(MPRIS_PLAYER_PREFIX):
                self._add_player(name)

        for name in list(self.players):
            self._get_playing_track(name)

        self._running = True

    def stop(self) -> None:
        for name in list(self.players):
            if name.startswith(MPRIS_PLAYER_PREFIX):
                self._remove_player(name)

    def _signal_name_owner_changed(
        self,
        _connection: Gio.DBusConnection,
        _sender_name: str,
        _object_path: str,
        _interface_name: str,
        _signal_name: str,
        parameters: tuple[str, str, str],
        *_user_data: Any,
    ):
        name, old_owner, new_owner = parameters
        if name.startswith(MPRIS_PLAYER_PREFIX):
            if new_owner and not old_owner:
                self._add_player(name)
            else:
                self._remove_player(name)

    def _add_player(self, name: str) -> None:
        """Set up a listener for music player signals"""
        log.info("%s appeared", name)

        if name in self.players:
            return

        assert self.connection is not None
        self.players[name] = self.connection.signal_subscribe(
            name,
            "org.freedesktop.DBus.Properties",
            "PropertiesChanged",
            "/org/mpris/MediaPlayer2",
            None,
            Gio.DBusSignalFlags.NONE,
            self._signal_received,
            name,
        )

    def _remove_player(self, name: str) -> None:
        log.info("%s vanished", name)
        if name in self.players:
            if self.connection is None:
                self._emit(None)
                return

            self.connection.signal_unsubscribe(self.players[name])
            self.players.pop(name)

            self._emit(None)

    def _signal_received(
        self,
        _connection: Gio.DBusConnection,
        _sender_name: str,
        _object_path: str,
        interface_name: str,
        _signal_name: str,
        parameters: tuple[str, str, str],
        *user_data: str,
    ):
        """Signal handler for PropertiesChanged event"""

        log.info("Signal received: %s - %s", interface_name, parameters)
        self._get_playing_track(user_data[0])

    @staticmethod
    def _get_music_info(properties: TrackProperties) -> TuneData | None:
        meta = properties.get("Metadata")
        if meta is None or not meta:
            return None

        status = properties.get("PlaybackStatus")
        if status is None or status == "Paused":
            return None

        title = meta.get("xesam:title")
        album = meta.get("xesam:album")
        # xesam:artist is always a list of strings if not None
        artist = meta.get("xesam:artist")
        if artist is not None:
            artist = ", ".join(artist)
        return TuneData(artist=artist, title=title, source=album)

    def _get_playing_track(self, name: str) -> None:
        """Return a TuneData for the currently playing
        song, or None if no song is playing"""
        try:
            proxy = Gio.DBusProxy.new_for_bus_sync(
                Gio.BusType.SESSION,
                Gio.DBusProxyFlags.NONE,
                None,
                name,
                "/org/mpris/MediaPlayer2",
                "org.freedesktop.DBus.Properties",
                None,
            )
        except GLib.Error as error:
            log.debug("Could not enable music listener: %s", error.message)
            return

        def proxy_call_finished(proxy: Gio.DBusProxy, res: Gio.AsyncResult) -> None:
            try:
                result = proxy.call_finish(res)
            except GLib.Error as error:
                log.debug("Could not enable music listener: %s", error.message)
                return

            assert result is not None
            info = self._get_music_info(result[0])
            if info is not None:
                self._emit(info)

        proxy.call(
            "GetAll",
            GLib.Variant("(s)", ("org.mpris.MediaPlayer2.Player",)),
            Gio.DBusCallFlags.NONE,
            -1,
            None,
            proxy_call_finished,
        )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only


from __future__ import annotations

from typing import Any

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject


class Singleton(type):
    _instances: dict[Any, Any] = {}

    def __call__(cls, *args: Any, **kwargs: Any):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class SettingsAction(Gio.SimpleAction):
    def simple_bind_property(
        self, target: GObject.Object, target_property: str
    ) -> None:
        self.bind_property(
            "state",
            target,
            target_property,
            GObject.BindingFlags.SYNC_CREATE,
            transform_to=self._transform_to_ptype,
        )

    def change_state(self, value: GLib.Variant | Any) -> None:
        if not isinstance(value, GLib.Variant):
            v_type = self.get_state_type()
            assert v_type is not None
            value = GLib.Variant(v_type.dup_string(), value)
        super().change_state(value)

    @staticmethod
    def _transform_to_ptype(binding: GObject.Binding, variant: GLib.Variant) -> str:
        return variant.unpack()



# Copyright (C) 2003-2017 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2004-2005 Vincent Hanquez <tab AT snarc.org>
# Copyright (C) 2005 Alex Podaras <bigpod AT gmail.com>
#                    Norman Rasmussen <norman AT rasmussen.co.za>
#                    Stéphan Kochen <stephan AT kochen.nl>
# Copyright (C) 2005-2006 Dimitur Kirov <dkirov AT gmail.com>
#                         Alex Mauer <hawke AT hawkesnest.net>
# Copyright (C) 2005-2007 Travis Shirk <travis AT pobox.com>
#                         Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2006 Junglecow J <junglecow AT gmail.com>
#                    Stefan Bethge <stefan AT lanpartei.de>
# Copyright (C) 2006-2008 Jean-Marie Traissard <jim AT lapin.org>
# Copyright (C) 2007 Lukas Petrovicky <lukas AT petrovicky.net>
#                    James Newton <redshodan AT gmail.com>
# Copyright (C) 2007-2008 Brendan Taylor <whateley AT gmail.com>
#                         Julien Pivotto <roidelapluie AT gmail.com>
#                         Stephan Erb <steve-e AT h3c.de>
# Copyright (C) 2008 Jonathan Schleifer <js-gajim AT webkeks.org>
# Copyright (C) 2016-2017 Emmanuel Gil Peyrot <linkmauve AT linkmauve.fr>
#                         Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import os
import shutil
import sys
from collections.abc import Callable
from datetime import datetime
from datetime import timedelta
from datetime import UTC
from pathlib import Path

from gi.repository import Adw
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk
from nbxmpp import JID
from nbxmpp.const import ConnectionProtocol
from nbxmpp.const import ConnectionType

import gajim
from gajim.common import app
from gajim.common import configpaths
from gajim.common import events
from gajim.common import ged
from gajim.common import idle
from gajim.common.application import CoreApplication
from gajim.common.const import GAJIM_FAQ_URI
from gajim.common.const import GAJIM_PRIVACY_POLICY_URI
from gajim.common.const import GAJIM_SUPPORT_JID
from gajim.common.const import GAJIM_WIKI_URI
from gajim.common.helpers import idle_add_once
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import ResourceContact

from gajim.gtk import structs
from gajim.gtk.alert import ConfirmationAlertDialog
from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.avatar import AvatarStorage
from gajim.gtk.const import ACCOUNT_ACTIONS
from gajim.gtk.const import ALWAYS_ACCOUNT_ACTIONS
from gajim.gtk.const import APP_ACTIONS
from gajim.gtk.const import FEATURE_ACCOUNT_ACTIONS
from gajim.gtk.const import MuteState
from gajim.gtk.const import ONLINE_ACCOUNT_ACTIONS
from gajim.gtk.shortcut_manager import ShortcutManager
from gajim.gtk.util.icons import get_icon_theme
from gajim.gtk.util.misc import open_file
from gajim.gtk.util.misc import open_uri
from gajim.gtk.util.window import get_app_window
from gajim.gtk.util.window import get_app_windows
from gajim.gtk.util.window import open_window

ActionListT = list[tuple[str, Callable[[Gio.SimpleAction, GLib.Variant], Any]]]


class GajimApplication(Adw.Application, CoreApplication):
    """Main class handling activation and command line."""

    def __init__(self):
        CoreApplication.__init__(self)
        flags = (
            Gio.ApplicationFlags.HANDLES_COMMAND_LINE
            | Gio.ApplicationFlags.CAN_OVERRIDE_APP_ID
        )
        Adw.Application.__init__(
            self, application_id=app.get_default_app_id(), flags=flags
        )

        # required to track screensaver state
        self.props.register_session = True

        self.add_main_option(
            "version",
            ord("V"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Show the application's version"),
        )

        self.add_main_option(
            "quiet",
            ord("q"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Show only critical errors"),
        )

        self.add_main_option(
            "separate",
            ord("s"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Separate profile files completely (even history database and plugins)"),
        )

        self.add_main_option(
            "verbose",
            ord("v"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Print XML stanzas and other debug information"),
        )

        self.add_main_option(
            "profile",
            ord("p"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING,
            _("Use defined profile in configuration directory"),
            "NAME",
        )

        self.add_main_option(
            "user-profile",
            ord("u"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING,
            _("Use a profile to run multiple Gajim instances"),
            "NAME",
        )

        self.add_main_option(
            "config-path",
            ord("c"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING,
            _("Set configuration directory"),
            "PATH",
        )

        self.add_main_option(
            "loglevel",
            ord("l"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING,
            _("Configure logging system"),
            "LEVEL",
        )

        self.add_main_option(
            "warnings",
            ord("w"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Show all warnings"),
        )

        self.add_main_option(
            "gdebug",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Sets an environment variable so GLib debug messages are printed"),
        )

        self.add_main_option(
            "cprofile",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Profile application with cprofile"),
        )

        self.add_main_option(
            "start-chat",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Start a new chat"),
        )

        self.add_main_option_entries(self._get_remaining_entry())

        self.connect("activate", self._on_activate)
        self.connect("handle-local-options", self._handle_local_options)
        self.connect("command-line", self._command_line)
        self.connect("shutdown", self._on_shutdown)
        self.connect("query-end", self._on_query_end)
        self.connect("window_added", self._on_window_added)

    @staticmethod
    def _get_remaining_entry():
        option = GLib.OptionEntry()
        # https://gitlab.gnome.org/GNOME/pygobject/-/issues/608
        option.arg = int(GLib.OptionArg.STRING_ARRAY)  # pyright: ignore
        option.arg_data = None
        option.arg_description = "[URI …]"
        option.flags = GLib.OptionFlags.NONE
        option.long_name = GLib.OPTION_REMAINING
        option.short_name = 0
        return [option]

    def _startup(self) -> None:
        if sys.platform in ("win32", "darwin"):
            # Changing the PANGOCAIRO_BACKEND is necessary on Windows/macOS
            # to render colored emoji glyphs
            os.environ["PANGOCAIRO_BACKEND"] = "fontconfig"

        app.ged.register_event_handler("db-migration", 0, self._on_db_migration)

        if not self._init_core():
            return

        icon_theme = get_icon_theme()
        icon_theme.add_search_path(str(configpaths.get("ICONS")))

        self.avatar_storage = AvatarStorage()
        self._shortcut_manager = ShortcutManager()

        app.load_css_config()

        from gajim.gtk.main import MainWindow

        main_window = MainWindow()

        from gajim.gtk import notification

        notification.init()

        idle.Monitor.set_interval(
            app.settings.get("autoawaytime") * 60, app.settings.get("autoxatime") * 60
        )

        from gajim.gtk.tray_icon import TrayIcon

        self.systray = TrayIcon()

        self._add_app_actions()
        accounts = app.settings.get_accounts()
        for account in accounts:
            self.add_account_actions(account)

        self._active_accounts_count = len(app.settings.get_active_accounts())

        self.update_app_actions_state()

        self.register_event("feature-discovered", ged.CORE, self._on_feature_discovered)

        main_window.init()

        if self._deprecated_options_used:
            migration_url = (
                "https://dev.gajim.org/gajim/gajim/-/wikis/Profile-Migration"
            )
            InformationAlertDialog(
                "Deprecation Warning",
                (
                    "The options <b>--profile</b> and <b>--separate</b> are deprecated "
                    "and will be removed in a future version of Gajim. Visit our "
                    f"<a href='{migration_url}'>Wiki</a> "
                    "to find the instructions on how to migrate."
                ),
                body_use_markup=True,
            )

        self._inhibit_cookie = self.inhibit(
            app.window, Gtk.ApplicationInhibitFlags.LOGOUT, _("Quitting…")
        )

        GLib.timeout_add(100, self._auto_connect)

    def _on_window_added(
        self, _application: GajimApplication, window: Gtk.Window
    ) -> None:
        self._shortcut_manager.install_shortcuts(window, ["app"])

    def start_shutdown(self) -> None:
        app.window.start_shutdown()
        CoreApplication._start_shutdown(self)

    def _shutdown_complete(self) -> None:
        CoreApplication._shutdown_complete(self)
        self.uninhibit(self._inhibit_cookie)
        self.quit()

    def _on_query_end(self, _application: GajimApplication) -> None:
        app.log("app").info("Session end signal received")
        # Use a idle call so Gtk can send QueryEndResponse on Linux
        idle_add_once(self.start_shutdown)

    def _on_shutdown(self, _application: GajimApplication) -> None:
        app.log("app").info("Shutdown Application")

    def _command_line(
        self, _application: GajimApplication, command_line: Gio.ApplicationCommandLine
    ) -> int:
        """Handles command line options not related to the startup of Gajim."""

        options = command_line.get_options_dict()

        remote_commands = [
            ("start-chat", GLib.Variant("as", ["", ""])),
        ]

        for cmd, parameter in remote_commands:
            if options.contains(cmd):
                self.activate_action(cmd, parameter)
                return 0

        remaining = options.lookup_value(
            GLib.OPTION_REMAINING, GLib.VariantType.new("as")
        )

        if remaining is not None:
            self.activate_action("handle-uri", remaining)
            return 0

        if not options.contains("is-first-startup"):
            # If no commands have been handled and it's not the first
            # startup, raise the application.
            self.activate()

        return 0

    def _handle_local_options(
        self, _application: Gtk.Application, options: GLib.VariantDict
    ) -> int:
        if options.contains("version"):
            print(gajim.__version__)
            return 0

        application_name = "Gajim"

        # --profile is deprecated with Gajim 2.3.0
        profile = options.lookup_value("profile")
        user_profile = options.lookup_value("user-profile")
        if user_profile is not None:
            if options.contains("separate"):
                print("--separate cannot be used with --user-profile")
                return 0

            if options.contains("profile"):
                print("--profile cannot be used with --user-profile")
                return 0

            configpaths.set_user_profile(user_profile.get_string())

        elif profile is not None:
            user_profile = profile
            configpaths.set_profile(user_profile.get_string())

        if user_profile is not None:
            # Incorporate user_profile name into application id
            # to have a single app instance for each user_profile.
            user_profile_str = user_profile.get_string()
            application_name = f"Gajim ({user_profile_str})"
            app_id = f"{self.get_application_id()}.{user_profile_str}"
            self.set_application_id(app_id)
            configpaths.set_user_profile(user_profile_str)

        GLib.set_application_name(application_name)

        try:
            self.register()
        except GLib.Error as error:
            quark = GLib.quark_try_string("g-io-error-quark")
            if error.matches(quark, Gio.IOErrorEnum.TIMED_OUT):
                sys.exit(
                    "Timeout reached for registering the application on DBus, "
                    "check if there is another broken Gajim process"
                )
            else:
                sys.exit(f"Unable to register application on DBus: {error}")

        if self.get_is_remote():
            print(
                "Gajim is already running. "
                "The primary instance will handle remote commands"
            )
            return -1

        self._deprecated_options_used = options.contains("profile") or options.contains(
            "separate"
        )

        options.insert_value("is-first-startup", GLib.Variant("b", True))
        self._core_command_line(options)
        self._startup()
        return -1

    def _on_activate(self, _application: Gtk.Application) -> None:
        app.window.show_window()

    def get_shortcut_manager(self) -> ShortcutManager:
        return self._shortcut_manager

    def _add_app_actions(self) -> None:
        for action in APP_ACTIONS:
            action_name, variant = action
            if variant is not None:
                variant = GLib.VariantType.new(variant)

            act = Gio.SimpleAction.new(action_name, variant)
            self.add_action(act)

        self._connect_app_actions()

    def _connect_app_actions(self) -> None:
        actions: ActionListT = [
            ("quit", self._on_quit_action),
            ("add-account", self._on_add_account_action),
            ("manage-proxies", self._on_manage_proxies_action),
            ("preferences", self._on_preferences_action),
            ("plugins", self._on_plugins_action),
            ("xml-console", self._on_xml_console_action),
            ("shortcuts", self._on_shortcuts_action),
            ("features", self._on_features_action),
            ("content", self._on_content_action),
            ("join-support-chat", self._on_join_support_chat),
            ("about", self._on_about_action),
            ("faq", self._on_faq_action),
            ("privacy-policy", self._on_privacy_policy_action),
            ("start-chat", self._on_new_chat_action),
            ("accounts", self._on_accounts_action),
            ("add-contact", self._on_add_contact_action),
            ("copy-text", self._on_copy_text_action),
            ("open-link", self._on_open_link_action),
            ("export-history", self._on_export_history_action),
            ("remove-history", self._on_remove_history_action),
            ("create-groupchat", self._on_create_groupchat_action),
            ("forget-groupchat", self._on_forget_groupchat_action),
            ("open-chat", self._on_open_chat_action),
            ("mute-chat", self._on_mute_chat_action),
            ("save-file-as", self._on_save_file_as),
            ("open-file", self._on_open_file_folder),
            ("open-folder", self._on_open_file_folder),
        ]

        for action in actions:
            action_name, func = action
            act = self.lookup_action(action_name)
            assert act is not None
            act.connect("activate", func)
            self.add_action(act)

    def add_account_actions(self, account: str) -> None:
        for action_name, type_ in ACCOUNT_ACTIONS:
            account_action_name = f"{account}-{action_name}"
            if self.has_action(account_action_name):
                raise ValueError("Trying to add action more than once")

            variant_type = None
            if type_ is not None:
                variant_type = GLib.VariantType.new(type_)
            act = Gio.SimpleAction.new(account_action_name, variant_type)
            act.set_enabled(action_name in ALWAYS_ACCOUNT_ACTIONS)
            self.add_action(act)

        self._connect_account_actions(account)

    def _connect_account_actions(self, account: str) -> None:
        actions = [
            ("add-contact", self._on_add_contact_account_action),
            ("services", self._on_services_action),
            ("profile", self._on_profile_action),
            ("pep-config", self._on_pep_config_action),
            ("open-event", self._on_open_event_action),
            ("mark-as-read", self._on_mark_as_read_action),
            ("block-contact", self._on_block_contact),
            ("remove-contact", self._on_remove_contact),
            ("execute-command", self._on_execute_command),
            ("subscription-accept", self._on_subscription_accept),
            ("subscription-deny", self._on_subscription_deny),
        ]

        for action_name, func in actions:
            account_action_name = f"{account}-{action_name}"
            act = self.lookup_action(account_action_name)
            assert act is not None
            act.connect("activate", func)

    def remove_account_actions(self, account: str) -> None:
        for action_name in self.list_actions():
            if action_name.startswith(f"{account}-"):
                self.remove_action(action_name)

    def set_action_state(self, action_name: str, state: bool) -> None:
        action = self.lookup_action(action_name)
        assert isinstance(action, Gio.SimpleAction)
        action.set_enabled(state)

    def set_account_actions_state(self, account: str, new_state: bool = False) -> None:
        for action_name in ONLINE_ACCOUNT_ACTIONS:
            self.set_action_state(f"{account}-{action_name}", new_state)

        # Disable all feature actions on disconnect
        if not new_state:
            for action_name in FEATURE_ACCOUNT_ACTIONS:
                self.set_action_state(f"{account}-{action_name}", new_state)

    def update_feature_actions_state(self, account: str) -> None:
        client = app.get_client(account)
        blocking_available = client.get_module("Blocking").supported

        self.set_action_state(f"{account}-block-contact", blocking_available)

    def update_app_actions_state(self) -> None:
        active_accounts = bool(app.get_connected_accounts(exclude_local=True))
        self.set_action_state("create-groupchat", active_accounts)

        enabled_accounts = bool(app.settings.get_active_accounts())
        self.set_action_state("start-chat", enabled_accounts)

    def _on_feature_discovered(self, event: events.FeatureDiscovered) -> None:
        self.update_feature_actions_state(event.account)

    def create_account(
        self,
        account: str,
        address: JID,
        password: str,
        proxy_name: str | None,
        custom_host: tuple[str, ConnectionProtocol, ConnectionType] | None,
        anonymous: bool = False,
    ) -> None:
        CoreApplication.create_account(
            self, account, address, password, proxy_name, custom_host, anonymous
        )

        app.css_config.refresh()

        # Action must be added before account window is updated
        self.add_account_actions(account)

    @property
    def multi_account_mode(self) -> bool:
        # Store count in variable to reduce db queries
        return self._active_accounts_count > 1

    def enable_account(self, account: str) -> None:
        CoreApplication.enable_account(self, account)

        self.update_app_actions_state()

        self._active_accounts_count = len(app.settings.get_active_accounts())

    def disable_account(self, account: str) -> None:
        for win in get_app_windows(account):
            # Close all account specific windows, except the RemoveAccount
            # dialog. It shows if the removal was successful.
            if type(win).__name__ == "RemoveAccount":
                continue
            win.close()

        CoreApplication.disable_account(self, account)

        self.update_app_actions_state()

        self._active_accounts_count = len(app.settings.get_active_accounts())

    def remove_account(self, account: str) -> None:
        CoreApplication.remove_account(self, account)

        self.remove_account_actions(account)

    def _on_db_migration(self, _event: events.DBMigration) -> None:
        open_window("DBMigration")

        context = GLib.MainContext.default()
        while context.pending():
            context.iteration(may_block=False)

    # Action Callbacks

    @staticmethod
    def _on_add_contact_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        jid = param.get_string() or None
        if jid is not None:
            jid = JID.from_string(jid)
        open_window("AddContact", account=None, jid=jid)

    @staticmethod
    def _on_preferences_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        open_window("Preferences")

    @staticmethod
    def _on_plugins_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        open_window("Preferences").show_page("plugins")

    @staticmethod
    def _on_accounts_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        window = open_window("Preferences")

        account = param.get_string()
        if account:
            window.show_page(account)

    @staticmethod
    def _on_quit_action(_action: Gio.SimpleAction, _param: GLib.Variant | None) -> None:
        app.app.start_shutdown()

    @staticmethod
    def _on_new_chat_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        jid, initial_message = param.get_strv()
        open_window(
            "StartChatDialog",
            initial_jid=jid or None,
            initial_message=initial_message or None,
        )

    @staticmethod
    def _on_profile_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        account = param.get_string()
        open_window("ProfileWindow", account=account)

    @staticmethod
    def _on_services_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        account = param.get_string()
        open_window("ServiceDiscoveryWindow", account=account, address_entry=True)

    @staticmethod
    def _on_create_groupchat_action(
        _action: Gio.SimpleAction, param: GLib.Variant
    ) -> None:
        account = param.get_string()
        open_window("CreateGroupchatWindow", account=account or None)

    @structs.actionmethod
    def _on_add_contact_account_action(
        self, _action: Gio.SimpleAction, params: structs.AccountJidParam
    ) -> None:
        open_window("AddContact", account=params.account, jid=params.jid)

    @staticmethod
    def _on_add_account_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        open_window("AccountWizard")

    @structs.actionmethod
    def _on_block_contact(
        self, _action: Gio.SimpleAction, params: structs.AccountJidParam
    ) -> None:
        app.window.block_contact(params.account, params.jid)

    def _on_remove_contact(
        self, _action: Gio.SimpleAction, param: GLib.Variant
    ) -> None:
        account, jid = param.unpack()
        app.window.remove_contact(account, JID.from_string(jid))

    def _on_execute_command(
        self, _action: Gio.SimpleAction, param: GLib.Variant
    ) -> None:
        account, jids = param.unpack()
        open_window("AdHocCommands", account=account, jids=jids)

    @staticmethod
    @structs.actionfunction
    def _on_subscription_accept(
        _action: Gio.SimpleAction, params: structs.SubscriptionAcceptParam
    ) -> None:
        client = app.get_client(params.account)
        client.get_module("Presence").subscribed(params.jid)
        contact = client.get_module("Contacts").get_contact(params.jid)
        assert isinstance(contact, BareContact)
        if not contact.is_in_roster:
            open_window(
                "AddContact",
                account=params.account,
                jid=params.jid,
                nick=params.nickname or contact.name,
            )

    @staticmethod
    @structs.actionfunction
    def _on_subscription_deny(
        _action: Gio.SimpleAction, params: structs.AccountJidParam
    ) -> None:
        client = app.get_client(params.account)
        client.get_module("Presence").unsubscribed(params.jid)

    @staticmethod
    def _on_pep_config_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        account = param.get_string()
        open_window("PEPConfig", account=account)

    @staticmethod
    def _on_xml_console_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        open_window("DebugConsoleWindow")

    @staticmethod
    def _on_manage_proxies_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        open_window("ManageProxies")

    @staticmethod
    def _on_content_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        open_uri(GAJIM_WIKI_URI)

    @staticmethod
    def _on_join_support_chat(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        accounts = app.settings.get_active_accounts()
        if len(accounts) == 1:
            app.window.show_add_join_groupchat(accounts[0], GAJIM_SUPPORT_JID)
            return
        open_window("StartChatDialog", initial_jid=GAJIM_SUPPORT_JID)

    @staticmethod
    def _on_faq_action(_action: Gio.SimpleAction, _param: GLib.Variant | None) -> None:
        open_uri(GAJIM_FAQ_URI)

    @staticmethod
    def _on_privacy_policy_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        open_uri(GAJIM_PRIVACY_POLICY_URI)

    @staticmethod
    def _on_shortcuts_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        open_window("Preferences").show_page("shortcuts")

    @staticmethod
    def _on_features_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        open_window("Features")

    @staticmethod
    def _on_about_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        app.window.about_dialog.present()

    # @staticmethod
    # TODO Jingle FT
    # def _on_file_transfer_action(_action: Gio.SimpleAction,
    #                              _param: GLib.Variant | None) -> None:

    #     ft = app.interface.instances['file_transfers']
    #     if ft.window.get_property('visible'):
    #         ft.window.present()
    #     else:
    #         ft.window.show_all()

    @staticmethod
    @structs.actionfunction
    def _on_open_event_action(
        _action: Gio.SimpleAction, params: structs.OpenEventActionParams
    ) -> None:
        if params.type in ("connection-failed", "server-shutdown"):
            app.window.show_account_page(params.account)

        elif params.type in (
            "subscription-request",
            "muc-invitation",
            "reaction-updated",
        ):
            app.window.show_activity_page(params.context_id)

        elif params.type in ("incoming-message", "incoming-call", "file-transfer"):
            assert params.jid
            jid = JID.from_string(params.jid)
            app.window.select_chat(params.account, jid)

        app.window.present()

    @structs.actionmethod
    def _on_mark_as_read_action(
        self, _action: Gio.SimpleAction, params: structs.AccountJidParam
    ) -> None:
        app.window.mark_as_read(params.account, params.jid)

    @staticmethod
    def _on_open_link_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        open_uri(param.get_string())

    @staticmethod
    def _on_copy_text_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        app.window.get_clipboard().set(param.get_string())

    @staticmethod
    def _on_open_chat_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        account, jid = param.get_strv()
        app.window.start_chat_from_jid(account, jid)

    @staticmethod
    @structs.actionfunction
    def _on_mute_chat_action(
        _action: Gio.SimpleAction, params: structs.MuteContactParam
    ) -> None:
        client = app.get_client(params.account)
        contact = client.get_module("Contacts").get_contact(params.jid)
        assert not isinstance(contact, ResourceContact)

        if params.state == MuteState.UNMUTE:
            contact.settings.set("mute_until", None)
            return

        until = datetime.now(UTC) + timedelta(minutes=params.state)
        contact.settings.set("mute_until", until.isoformat())

    @staticmethod
    @structs.actionfunction
    def _on_export_history_action(
        _action: Gio.SimpleAction, params: structs.ExportHistoryParam
    ) -> None:
        open_window("HistoryExport", account=params.account, jid=params.jid)

    @staticmethod
    @structs.actionfunction
    def _on_remove_history_action(
        _action: Gio.SimpleAction, params: structs.AccountJidParam
    ) -> None:
        def _on_response() -> None:
            app.storage.archive.remove_history_for_jid(params.account, params.jid)

            app.window.clear_chat_list_row(params.account, params.jid)
            control = app.window.get_control()
            if not control.is_chat_active(params.account, params.jid):
                return

            control.reset_view()

        ConfirmationAlertDialog(
            _("Remove Chat History?"),
            _("Do you want to remove all chat history for this chat?"),
            confirm_label=_("_Remove"),
            appearance="destructive",
            callback=_on_response,
        )

    @staticmethod
    @structs.actionfunction
    def _on_forget_groupchat_action(
        _action: Gio.SimpleAction, params: structs.AccountJidParam
    ) -> None:
        def _on_response() -> None:
            window = get_app_window("StartChatDialog")
            if window is not None:
                window.remove_row(params.account, params.jid)

            client = app.get_client(params.account)
            client.get_module("MUC").leave(params.jid)
            client.get_module("Bookmarks").remove(params.jid)

            app.storage.archive.remove_history_for_jid(params.account, params.jid)

        ConfirmationAlertDialog(
            _("Forget this Group Chat?"),
            _("Do you want to remove this chat including its chat history?"),
            confirm_label=_("_Remove"),
            appearance="destructive",
            callback=_on_response,
        )

    def _on_save_file_as(self, _action: Gio.SimpleAction, param: GLib.Variant) -> None:
        orig_path = Path(param.get_string())

        def _on_save_finished(
            file_dialog: Gtk.FileDialog, result: Gio.AsyncResult
        ) -> None:
            try:
                gfile = file_dialog.save_finish(result)
            except GLib.Error as e:
                if e.code == 2:
                    # User dismissed dialog, do nothing
                    return

                InformationAlertDialog(
                    _("Could Not Save File"),
                    _("Could not save file to selected directory."),
                )
                return

            path = gfile.get_path()
            assert path is not None
            target_path = Path(path)
            orig_ext = orig_path.suffix
            new_ext = target_path.suffix
            if orig_ext != new_ext:
                # Windows file chooser selects the full file name including
                # extension. Starting to type will overwrite the extension
                # as well. Restore the extension if it's lost.
                target_path = target_path.with_suffix(orig_ext)
            dirname = target_path.parent
            if not os.access(dirname, os.W_OK):
                InformationAlertDialog(
                    _("Directory Not Writable"),
                    _(
                        'Directory "%s" is not writable. '
                        "You do not have the proper permissions to "
                        "create files in this directory."
                    )
                    % dirname,
                )
                return

            try:
                shutil.copyfile(orig_path, target_path)
            except Exception as e:
                InformationAlertDialog(
                    _("Could Not Save File"),
                    _("There was an error while trying to save the file.\nError: %s.")
                    % e,
                )
                return

            app.settings.set("last_save_dir", str(target_path.parent))

            app.window.show_toast(
                Adw.Toast(
                    title=_("File saved"),
                    timeout=5,
                    button_label=_("Open Folder"),
                    action_name="app.open-folder",
                    action_target=GLib.Variant("s", str(target_path)),
                )
            )

        gfile = None
        last_dir = app.settings.get("last_save_dir")
        if last_dir:
            gfile = Gio.File.new_for_path(last_dir)

        dialog = Gtk.FileDialog(
            initial_folder=gfile,
            initial_name=orig_path.name,
        )
        dialog.save(app.window, None, _on_save_finished)

    def _on_open_file_folder(
        self, action: Gio.SimpleAction, param: GLib.Variant
    ) -> None:
        open_file(
            Path(param.get_string()), show_in_folder=action.get_name() == "open-folder"
        )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

import functools
import importlib.metadata
import os
import platform
import sys

from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Soup
from packaging.requirements import Requirement
from packaging.specifiers import SpecifierSet

CURRENT_PYTHON_VERSION = platform.python_version()


def python_version(specifier_set: str) -> bool:
    spec = SpecifierSet(specifier_set)
    return CURRENT_PYTHON_VERSION in spec


def package_version(requirement: str) -> bool:
    req = Requirement(requirement)
    try:
        installed_version = importlib.metadata.version(req.name)
    except importlib.metadata.PackageNotFoundError:
        return False

    return req.specifier.contains(installed_version, prereleases=True)


def gi_package_version(requirement: str):
    req = Requirement(requirement)
    match req.name:
        case "GLib":
            version = get_glib_version()
        case "PyGObject":
            version = get_gobject_version()
        case "Soup":
            version = get_soup_version()
        case _:
            raise ValueError("Unknown lib name: %s" % req.name)

    return req.specifier.contains(version)


@functools.lru_cache(maxsize=1)
def get_os_info() -> str:
    info = "N/A"
    if sys.platform in ("win32", "darwin"):
        info = f"{platform.system()} {platform.release()}"

    elif sys.platform == "linux":
        try:
            import distro

            info = distro.name(pretty=True)
        except ImportError:
            info = platform.system()
    return info


def get_os_name() -> str:
    if sys.platform in ("win32", "darwin"):
        return platform.system()
    if os.name == "posix":
        try:
            import distro

            return distro.name(pretty=True)
        except ImportError:
            return platform.system()
    return ""


def get_os_version() -> str:
    if sys.platform in ("win32", "darwin"):
        return platform.version()
    if os.name == "posix":
        try:
            import distro

            return distro.version(pretty=True)
        except ImportError:
            return platform.release()
    return ""


def get_gobject_version() -> str:
    return ".".join(map(str, GObject.pygobject_version))  # pyright: ignore


def get_glib_version() -> str:
    return ".".join(
        map(str, [GLib.MAJOR_VERSION, GLib.MINOR_VERSION, GLib.MICRO_VERSION])
    )


def get_soup_version() -> str:
    return ".".join(
        map(
            str,
            [
                Soup.get_major_version(),
                Soup.get_minor_version(),
                Soup.get_micro_version(),
            ],
        )
    )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast

import logging
from pathlib import Path

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common import types
from gajim.common.const import Display
from gajim.common.const import IMAGE_MIME_TYPES
from gajim.common.helpers import load_file_async
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.multiprocess.thumbnail import create_thumbnail
from gajim.common.util.preview import get_icon_for_mime_type
from gajim.common.util.preview import guess_mime_type
from gajim.common.util.uri import get_file_path_from_uri

from gajim.gtk.builder import get_builder
from gajim.gtk.filechoosers import FileChooserButton
from gajim.gtk.resource_selector import ResourceSelector
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import gi_gui_package_version
from gajim.gtk.util.misc import iterate_listbox_children

PREVIEW_SIZE = 72

log = logging.getLogger("gajim.gtk.file_transfer_selector")


class FileTransferSelector(Gtk.Box, SignalManager):
    __gsignals__ = {
        "changed": (GObject.SignalFlags.RUN_LAST, None, (bool,)),
    }

    def __init__(self, contact: types.ChatContactT, method: str | None = None) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)

        self._contact = contact
        self._method = method or app.window.get_preferred_ft_method(contact)

        client = app.get_client(contact.account)
        self._max_http_file_size = client.get_module("HTTPUpload").max_file_size

        self._ui = get_builder("file_transfer_selector.ui", widgets=["stack"])
        self.append(self._ui.stack)

        last_dir = app.settings.get("last_send_dir") or None
        if last_dir is not None:
            last_dir = Path(last_dir)

        self._file_chooser_button = FileChooserButton(
            path=last_dir, multiple=True, label=_("Add Files…")
        )
        self._file_chooser_button.set_halign(Gtk.Align.CENTER)
        self._ui.box.append(self._file_chooser_button)

        self._connect(
            self._file_chooser_button, "path-picked", self._on_choose_files_clicked
        )

        # TODO Jingle FT
        # self._resource_selector = None

        # if isinstance(contact, BareContact):
        #     # No jingle file transfers in group chats
        #     self._resource_selector = ResourceSelector(
        #         contact,
        #         constraints=[Namespace.JINGLE_FILE_TRANSFER_5])
        #     self._connect(
        #         self._resource_selector,
        #         'selection-changed',
        #         self._on_resource_selection
        #     )
        #     self._ui.resource_box.prepend(
        #         self._resource_selector)

        #     self._ui.resource_instructions.set_text(
        #         _('%s is online with multiple devices.\n'
        #           'Choose the device you would like to send the '
        #           'files to.') % self._contact.name)

        drop_target = Gtk.DropTarget.new(Gdk.FileList, Gdk.DragAction.COPY)
        self._connect(drop_target, "accept", self._on_drop_accept)
        self._connect(drop_target, "drop", self._on_file_drop)
        self.add_controller(drop_target)

    def do_unroot(self) -> None:
        Gtk.Box.do_unroot(self)
        self._disconnect_all()
        del self._file_chooser_button
        app.check_finalize(self)

    def _is_over_max_http_file_size(self, path: Path) -> bool:
        if self._max_http_file_size is None:
            return False

        if not path.exists():
            # Non-existing files are handled in the send_file method
            return False

        return path.stat().st_size > self._max_http_file_size

    def transfer_resource_required(self) -> bool:
        if not isinstance(self._contact, BareContact):
            # No jingle file transfer for group chats
            return False

        if self._method is None:
            log.error("No file transfer method available")
            return False

        current_page = self._ui.stack.get_visible_child_name()
        if self._method == "httpupload":
            for path in self._get_file_paths():
                if self._is_over_max_http_file_size(path):
                    if current_page == "resource-selection":
                        return False

                    self._ui.stack.set_visible_child_name("resource-selection")
                    self.emit("changed", False)
                    return True
            return False

        if current_page == "resource-selection":
            return False

        self._ui.stack.set_visible_child_name("resource-selection")
        self.emit("changed", False)
        return True

    def get_catalog(self) -> list[tuple[Path, str, JID]]:
        # catalog: list[(file Path, transfer method, recipient JID)]
        catalog: list[tuple[Path, str, JID]] = []

        file_paths = self._get_file_paths()
        for path in file_paths:
            # if (self._method == 'jingle' or
            #         self._is_over_max_http_file_size(path)):
            #     assert self._resource_selector is not None
            #     item = (path, 'jingle', self._resource_selector.get_jid())
            if self._method == "httpupload":
                item = (path, "httpupload", self._contact.jid)
            else:
                # No file transfer method available
                continue

            catalog.append(item)

        return catalog

    def add_files(self, uris: list[str]) -> None:
        for uri in uris:
            path = get_file_path_from_uri(uri)
            if path is None:
                self._add_warning_message("Could not add %s" % uri)
                continue

            size_warning = bool(
                self._method == "httpupload" and self._is_over_max_http_file_size(path)
            )
            jingle_warning = bool(self._method == "jingle")

            row = FileRow(path, size_warning, jingle_warning)
            self._connect(row, "removed", self._on_row_removed)
            self._ui.listbox.append(row)
            app.settings.set("last_send_dir", str(path.parent))

        self._on_files_changed()

    def _on_row_removed(self, row: FileRow) -> None:
        self._disconnect_object(row)
        self._ui.listbox.remove(row)
        self._on_files_changed()

    def _add_warning_message(self, msg: str) -> None:
        log.warning(msg)  # TODO: replace with UI

    def _get_file_paths(self) -> list[Path]:
        paths: list[Path] = []
        for row in cast(list[FileRow], iterate_listbox_children(self._ui.listbox)):
            paths.append(row.file_path)

        return paths

    def _on_drop_accept(self, _target: Gtk.DropTarget, drop: Gdk.Drop) -> bool:
        # DND on X11 freezes due to a GTK bug:
        # https://dev.gajim.org/gajim/gajim/-/issues/12313
        if app.is_display(Display.X11) and not gi_gui_package_version("Gtk>=4.20.1"):
            return False

        formats = drop.get_formats()
        return bool(formats.contain_gtype(Gdk.FileList))

    def _on_file_drop(
        self, _target: Gtk.DropTarget, value: Gdk.FileList | None, _x: float, _y: float
    ) -> bool:
        if value is None:
            log.debug("Drop received, but value is None")
            return False

        files = value.get_files()
        if not files:
            return False

        self.add_files([file.get_uri() for file in files])
        return True

    def _on_files_changed(self) -> None:
        file_paths = self._get_file_paths()
        if not file_paths:
            self.emit("changed", False)
            return

        # if self._method == 'httpupload' and self._contact.is_groupchat:
        #     # Enforce HTTPUpload file size limit for group chats
        #     for path in file_paths:
        #         if self._is_over_max_http_file_size(path):
        #             self.emit('changed', False)
        #             return

        if self._method == "httpupload":
            # Enforce HTTPUpload file size limit
            for path in file_paths:
                if self._is_over_max_http_file_size(path):
                    self.emit("changed", False)
                    return

        self.emit("changed", True)

    def _on_resource_selection(self, _selector: ResourceSelector, state: bool) -> None:
        self.emit("changed", state)

    def _on_choose_files_clicked(
        self, _button: FileChooserButton, paths: list[Path]
    ) -> None:
        self.add_files([p.as_uri() for p in paths])
        self._file_chooser_button.reset()


class FileRow(Gtk.ListBoxRow, SignalManager):
    __gsignals__ = {
        "removed": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, path: Path, size_warning: bool, jingle_warning: bool) -> None:
        Gtk.ListBoxRow.__init__(self)
        SignalManager.__init__(self)

        self.file_path = path

        self._ui = get_builder("file_transfer_selector.ui", widgets=["file_box"])
        self.set_child(self._ui.file_box)

        self._connect(self._ui.remove_file_button, "clicked", self._on_remove_clicked)

        self._ui.file_name_label.set_text(path.name)

        if app.settings.get("use_kib_mib"):
            units = GLib.FormatSizeFlags.IEC_UNITS
        else:
            units = GLib.FormatSizeFlags.DEFAULT

        file_size_string = GLib.format_size_full(path.stat().st_size, units)
        self._ui.file_size_label.set_text(file_size_string)

        # if size_warning:
        #     self._ui.warning_label.set_text(
        #         _('File too big, will use direct transfer (not encrypted)'))
        #     self._ui.warning_label.set_visible(True)

        if size_warning:
            self._ui.warning_label.set_text(_("File too big"))
            self._ui.warning_label.set_visible(True)

        # if jingle_warning:
        #     self._ui.warning_label.set_text(
        #         _('Direct transfer (not encrypted)'))
        #     self._ui.warning_label.set_visible(True)

        self._ui.preview_image_box.set_size_request(PREVIEW_SIZE, -1)
        self._set_preview(path)

    def do_unroot(self) -> None:
        Gtk.ListBoxRow.do_unroot(self)
        self._disconnect_all()
        app.check_finalize(self)

    def _set_preview(self, path: Path) -> None:
        mime_type = guess_mime_type(path)
        if mime_type not in IMAGE_MIME_TYPES:
            self._set_icon_for_mime_type(mime_type)
            return

        load_file_async(self.file_path, self._on_load_finished, mime_type)

    def _on_load_finished(
        self, data: bytes | None, _error: GLib.Error | None, mime_type: Any
    ) -> None:
        if data is None:
            self._set_icon_for_mime_type(mime_type)
            return

        try:
            thumbnail_bytes, _metadata = create_thumbnail(
                data, None, PREVIEW_SIZE, mime_type
            )
        except Exception as error:
            log.error("Failed to create thumbnail: %s", error)
            self._set_icon_for_mime_type(mime_type)
            return

        texture = Gdk.Texture.new_from_bytes(GLib.Bytes.new(thumbnail_bytes))

        self._ui.preview_image.set_from_paintable(texture)

    def _set_icon_for_mime_type(self, mime_type: str) -> None:
        icon = get_icon_for_mime_type(mime_type)
        self._ui.preview_image.set_from_gicon(icon)

    def _on_remove_clicked(self, _button: Gtk.Button) -> None:
        self.emit("removed")



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import cairo
import nbxmpp
from gi.repository import Adw
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common.const import ARTISTS
from gajim.common.const import DEVELOPERS
from gajim.common.const import MAINTAINERS
from gajim.common.const import THANKS
from gajim.common.i18n import _
from gajim.common.util.app import get_extended_app_version
from gajim.common.util.version import get_glib_version
from gajim.common.util.version import get_gobject_version
from gajim.common.util.version import get_soup_version

from gajim.gtk.util.misc import get_adw_version
from gajim.gtk.util.misc import get_gtk_version


class AboutDialog:
    def __init__(self) -> None:
        self._dialog = None

    def present(self) -> None:
        if self._dialog is None:
            self._dialog = self._get_dialog()
        self._dialog.present()

    def _get_dialog(self) -> Adw.AboutDialog:
        dialog = Adw.AboutDialog(
            application_name="Gajim",
            application_icon="gajim",
            version=get_extended_app_version(),
            copyright="Copyright © 2003-2026 Gajim Team",
            license_type=Gtk.License.GPL_3_0_ONLY,
            website="https://gajim.org/",
            issue_url="https://dev.gajim.org/gajim/gajim/-/issues/",
            developer_name="\n".join(MAINTAINERS),
            developers=MAINTAINERS + DEVELOPERS,
            designers=ARTISTS,
            debug_info=self._get_debug_info(),
            debug_info_filename="gajim_version_info.txt",
            translator_credits=_("translator-credits"),
        )
        dialog.add_acknowledgement_section(_("Thanks"), THANKS)
        dialog.add_acknowledgement_section(
            _("Packages"), [_("Thanks to all the package maintainers.")]
        )
        dialog.connect("closed", self._on_dialog_closed)
        return dialog

    def _get_debug_info(self) -> str:
        debug_info = [
            _("GTK Version: %s") % get_gtk_version(),
            _("Adw Version: %s") % get_adw_version(),
            _("GLib Version: %s") % get_glib_version(),
            _("Pango Version: %s") % Pango.version_string(),
            _("PyGObject Version: %s") % get_gobject_version(),
            _("cairo Version: %s") % cairo.cairo_version_string(),
            _("pycairo Version: %s") % cairo.version,
            _("python-nbxmpp Version: %s") % nbxmpp.__version__,
            _("libsoup Version: %s") % get_soup_version(),
        ]
        return "\n".join(debug_info)

    def _on_dialog_closed(self, _dialog: AboutDialog) -> None:
        self._dialog = None



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import overload
from typing import TypedDict

import logging

from gi.repository import GObject
from gi.repository import Gtk
from nbxmpp import JID
from nbxmpp import Namespace
from nbxmpp.errors import BaseError
from nbxmpp.errors import is_error
from nbxmpp.errors import StanzaError
from nbxmpp.structs import DiscoInfo

from gajim.common import app
from gajim.common import types
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.util import as_task
from gajim.common.util.jid import validate_jid
from gajim.common.util.user_strings import get_subscription_request_msg

from gajim.gtk.assistant import Assistant
from gajim.gtk.assistant import AssistantErrorPage
from gajim.gtk.assistant import AssistantPage
from gajim.gtk.dropdown import GajimDropDown
from gajim.gtk.groupchat_info import GroupChatInfoScrolled
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.util.window import open_window

log = logging.getLogger("gajim.gtk.add_contact")


class SubscriptionData(TypedDict):
    message: str
    groups: list[str]
    auto_auth: bool


class AddContact(Assistant):
    def __init__(
        self,
        account: str | None = None,
        jid: JID | None = None,
        nick: str | None = None,
    ):
        Assistant.__init__(self)
        self.account = account
        self.jid = jid
        self._nick = nick

        self._result: DiscoInfo | BaseError | StanzaError | None = None

        self.add_button("next", _("Next"), complete=True, css_class="suggested-action")
        self.add_button("back", _("Back"))
        self.add_button("add", _("Add Contact"), css_class="suggested-action")
        self.add_button("join", _("Join…"), css_class="suggested-action")

        self.add_pages(
            {
                "address": AddContactAddressPage(account, jid),
                "error": AddContactErrorPage(),
                "contact": AddContactContactPage(),
                "groupchat": AddContactGroupChatPage(),
                "gateway": AddContactGatewayPage(),
            }
        )

        progress = self.add_default_page("progress")
        progress.set_title(_("Gathering information…"))
        progress.set_text(_("Trying to gather information on this address…"))

        self._connect(self, "button-clicked", self._on_button_clicked)
        self.show_first_page()

    @overload
    def get_page(self, name: Literal["address"]) -> AddContactAddressPage: ...

    @overload
    def get_page(self, name: Literal["error"]) -> AddContactErrorPage: ...

    @overload
    def get_page(self, name: Literal["contact"]) -> AddContactContactPage: ...

    @overload
    def get_page(self, name: Literal["groupchat"]) -> AddContactGroupChatPage: ...

    @overload
    def get_page(self, name: Literal["gateway"]) -> AddContactGatewayPage: ...

    def get_page(self, name: str) -> AssistantPage:
        return self._pages[name]

    def _on_button_clicked(self, _assistant: Assistant, button_name: str) -> None:
        page = self.get_current_page()
        address_page = self.get_page("address")
        account, _ = address_page.get_account_and_jid()

        if button_name == "next":
            self._start_disco()
            return

        if button_name == "back":
            self.show_page("address", Gtk.StackTransitionType.SLIDE_RIGHT)
            address_page.focus()
            return

        if button_name == "add":
            assert isinstance(self._result, DiscoInfo | StanzaError)
            assert isinstance(self._result.jid, JID)
            client = app.get_client(account)

            if page == "contact":
                contact_page = self.get_page("contact")
                data = contact_page.get_subscription_data()
                client.get_module("Presence").subscribe(
                    self._result.jid,
                    msg=data["message"],
                    groups=data["groups"],
                    auto_auth=data["auto_auth"],
                    name=self._nick,
                )

            elif page == "gateway":
                assert isinstance(self._result, DiscoInfo)
                client.get_module("Presence").subscribe(
                    self._result.jid, name=self._result.gateway_name, auto_auth=True
                )

            else:
                raise ValueError("No handler for page: %s" % page)

            app.window.add_chat(account, self._result.jid, "chat", select=True)
            self.close()
            return

        if button_name == "join":
            _, jid = address_page.get_account_and_jid()
            open_window("GroupchatJoin", account=account, jid=jid)
            self.close()

    def _start_disco(self) -> None:
        self._result = None
        self.show_page("progress", Gtk.StackTransitionType.SLIDE_LEFT)

        address_page = self.get_page("address")
        account, jid = address_page.get_account_and_jid()
        assert account is not None
        self._disco_info(account, jid)

    @as_task
    def _disco_info(self, account: str, address: str) -> Any:
        _task = yield  # noqa: F841, # pyright: ignore

        client = app.get_client(account)

        result = yield client.get_module("Discovery").disco_info(address, timeout=10)  # pyright: ignore
        if is_error(result):
            assert isinstance(result, BaseError)
            self._process_error(account, result)
            raise result

        log.info("Disco Info received: %s", address)
        assert isinstance(result, DiscoInfo)
        self._process_info(account, result)

    def _process_error(self, account: str, result: BaseError | StanzaError) -> None:
        log.debug("Error received: %s", result)
        self._result = result

        contact_conditions = [
            "service-unavailable",  # Prosody
            "subscription-required",  # ejabberd
        ]
        if isinstance(result, StanzaError) and result.condition in contact_conditions:
            # It seems to be a contact
            address_page = self.get_page("contact")
            address_page.prepare(account, result)
            self.show_page("contact", Gtk.StackTransitionType.SLIDE_LEFT)
        else:
            # TimeoutStanzaError is handled here
            error_page = self.get_page("error")
            error_page.set_text(result.get_text())
            self.show_page("error", Gtk.StackTransitionType.SLIDE_LEFT)

    def _process_info(self, account: str, result: DiscoInfo) -> None:
        log.debug("Info received: %s", result)
        self._result = result

        if result.is_gateway:
            gateway_page = self.get_page("gateway")
            gateway_page.prepare(account, result)
            self.show_page("gateway", Gtk.StackTransitionType.SLIDE_LEFT)
            return

        if result.is_muc:
            for identity in result.identities:
                assert result.jid is not None
                if identity.type == "text" and result.jid.is_domain:
                    # It's a group chat component advertising
                    # category 'conference'
                    error_page = self.get_page("error")
                    error_page.set_text(
                        _("This address does not seem to offer any gateway service.")
                    )
                    self.show_page("error")
                    return

                if identity.type == "irc" and result.jid.is_domain:
                    # It's an IRC gateway advertising category 'conference'
                    gateway_page = self.get_page("gateway")
                    gateway_page.prepare(account, result)
                    self.show_page("gateway", Gtk.StackTransitionType.SLIDE_LEFT)
                    return

            groupchat_page = self.get_page("groupchat")
            groupchat_page.prepare(account, result)
            self.show_page("groupchat", Gtk.StackTransitionType.SLIDE_LEFT)
            return

        contact_page = self.get_page("contact")
        contact_page.prepare(account, result)
        self.show_page("contact", Gtk.StackTransitionType.SLIDE_LEFT)


@Gtk.Template(string=get_ui_string("add_contact/address_page.ui"))
class AddContactAddressPage(AssistantPage):
    __gtype_name__ = "AddContactAddressPage"

    _account_box: Gtk.Box = Gtk.Template.Child()
    _dropdown: GajimDropDown[str] = Gtk.Template.Child()
    _address_entry: Gtk.Entry = Gtk.Template.Child()

    def __init__(self, account: str | None, jid: JID | None) -> None:
        AssistantPage.__init__(self)
        self.title = _("Add Contact")

        self._account = account
        self._jid = jid

        self._connect(self._dropdown, "notify::selected", self._on_account_selected)
        self._connect(self._address_entry, "changed", self._set_complete)

        accounts = app.get_enabled_accounts_with_labels(connected_only=True)
        self._dropdown.set_data(dict(accounts))

        self._account_box.set_visible(len(accounts) > 1)

        if account is not None:
            self._dropdown.select_key(account)

        if jid is not None:
            self._address_entry.set_text(str(jid))

        self._set_complete()

    def get_visible_buttons(self) -> list[str]:
        return ["next"]

    def get_default_button(self) -> str:
        return "next"

    def get_account_and_jid(self) -> tuple[str, str]:
        assert self._account is not None
        return self._account, self._address_entry.get_text()

    def focus(self) -> None:
        self._address_entry.grab_focus()

    def _on_account_selected(
        self, dropdown: GajimDropDown[str], _param: GObject.ParamSpec
    ) -> None:
        account = dropdown.get_selected_key()
        self._account = account
        self._set_complete()

    def _show_icon(self, show: bool) -> None:
        icon = "lucide-circle-alert-symbolic" if show else None
        self._address_entry.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, icon
        )

    def _set_complete(self, *args: Any) -> None:
        if self._account is None:
            self.complete = False
            self.update_page_complete()
            return

        address = self._address_entry.get_text()
        is_self = bool(address == app.get_jid_from_account(self._account))
        if is_self:
            self._show_icon(True)
            self._address_entry.set_icon_tooltip_text(
                Gtk.EntryIconPosition.SECONDARY,
                _("You cannot add yourself to your contact list."),
            )
            self.complete = False
            self.update_page_complete()
            return

        client = app.get_client(self._account)
        for contact in client.get_module("Roster").iter_contacts():
            if address == str(contact.jid):
                self._show_icon(True)
                self._address_entry.set_icon_tooltip_text(
                    Gtk.EntryIconPosition.SECONDARY,
                    _("%s is already in your contact list") % address,
                )
                self.complete = False
                self.update_page_complete()
                return

        self.complete = self._validate_address(address)
        self.update_page_complete()

    def _validate_address(self, address: str) -> bool:
        if not address:
            self._show_icon(False)
            return False

        try:
            jid = validate_jid(address)
            if jid.resource:
                raise ValueError
        except ValueError:
            self._show_icon(True)
            self._address_entry.set_icon_tooltip_text(
                Gtk.EntryIconPosition.SECONDARY, _("Invalid Address")
            )
            return False

        if jid.localpart is None:
            self._show_icon(True)
            self._address_entry.set_icon_tooltip_text(
                Gtk.EntryIconPosition.SECONDARY,
                _("'@' is missing in address. Are you sure this is correct?"),
            )
            return True

        self._show_icon(False)
        return True


class AddContactErrorPage(AssistantErrorPage):
    def __init__(self) -> None:
        AssistantErrorPage.__init__(self)
        self.set_title(_("Add Contact"))
        self.set_heading(_("An Error Occurred"))

    def get_visible_buttons(self) -> list[str]:
        return ["back"]

    def get_default_button(self) -> str:
        return "back"


@Gtk.Template(string=get_ui_string("add_contact/contact_page.ui"))
class AddContactContactPage(AssistantPage):
    __gtype_name__ = "AddContactContactPage"

    _contact_grid: Gtk.Grid = Gtk.Template.Child()
    _status_switch: Gtk.Switch = Gtk.Template.Child()
    _group_combo: Gtk.ComboBoxText = Gtk.Template.Child()
    _message_entry: Gtk.Entry = Gtk.Template.Child()
    _contact_info_button: Gtk.Button = Gtk.Template.Child()

    def __init__(self) -> None:
        AssistantPage.__init__(self)
        self.title = _("Add Contact")

        self._result: DiscoInfo | BaseError | None = None
        self._account: str | None = None
        self._contact: types.BareContact | None = None

        entry = self._group_combo.get_child()
        assert isinstance(entry, Gtk.Entry)
        entry.set_activates_default(True)
        entry.set_placeholder_text(_("Choose a group…"))

        self._connect(self._contact_info_button, "clicked", self._on_info_clicked)

    def get_visible_buttons(self) -> list[str]:
        return ["back", "add"]

    def get_default_button(self) -> str:
        return "add"

    def prepare(self, account: str, result: DiscoInfo | StanzaError):
        self._result = result
        self._account = account

        client = app.get_client(account)
        assert result.jid is not None
        contact = client.get_module("Contacts").get_contact(result.jid)
        assert isinstance(contact, BareContact)
        self._contact = contact

        self._update_groups(account)
        self._message_entry.set_text(get_subscription_request_msg(account))
        self._contact_grid.set_sensitive(True)

    def _update_groups(self, account: str) -> None:
        model = self._group_combo.get_model()
        assert isinstance(model, Gtk.ListStore)
        model.clear()
        client = app.get_client(account)
        for group in client.get_module("Roster").get_groups():
            self._group_combo.append_text(group)

    def _on_info_clicked(self, _button: Gtk.Button) -> None:
        open_window("ContactInfo", account=self._account, contact=self._contact)

    def get_subscription_data(self) -> SubscriptionData:
        entry = self._group_combo.get_child()
        assert isinstance(entry, Gtk.Entry)
        group = entry.get_text()
        groups = [group] if group else []
        return {
            "message": self._message_entry.get_text(),
            "groups": groups,
            "auto_auth": self._status_switch.get_active(),
        }


@Gtk.Template(string=get_ui_string("add_contact/gateway_page.ui"))
class AddContactGatewayPage(AssistantPage):
    __gtype_name__ = "AddContactGatewayPage"

    _gateway_image: Gtk.Image = Gtk.Template.Child()
    _gateway_label: Gtk.Label = Gtk.Template.Child()
    _register_button: Gtk.Button = Gtk.Template.Child()
    _commands_button: Gtk.Button = Gtk.Template.Child()

    def __init__(self) -> None:
        AssistantPage.__init__(self)
        self.title = _("Service Gateway")

        self._account: str | None = None
        self._result: DiscoInfo | None = None

        self._connect(self._register_button, "clicked", self._on_register_clicked)
        self._connect(self._commands_button, "clicked", self._on_command_clicked)

    def get_visible_buttons(self) -> list[str]:
        return ["back", "add"]

    def get_default_button(self) -> str:
        return "add"

    def prepare(self, account: str, result: DiscoInfo) -> None:
        self._account = account
        self._result = result

        icon_name = None
        if result.is_gateway:
            if result.gateway_type == "sms":
                icon_name = "gajim-agent-sms"
            if result.gateway_type == "irc":
                icon_name = "gajim-agent-irc"
            self._gateway_image.set_from_icon_name(icon_name)
            gateway_name = result.gateway_name or str(self._result.jid)
            if not result.gateway_type:
                self._gateway_label.set_text(gateway_name)
            else:
                self._gateway_label.set_text(
                    f"{gateway_name} ({result.gateway_type.upper()})"
                )
        else:
            identity_name = ""
            identity_type = ""
            for identity in result.identities:
                if identity.type == "sms":
                    icon_name = "gajim-agent-sms"
                    identity_name = identity.name or str(self._result.jid)
                    identity_type = identity.type
                if identity.type == "irc":
                    icon_name = "gajim-agent-irc"
                    identity_name = identity.name or str(self._result.jid)
                    identity_type = identity.type
            self._gateway_image.set_from_icon_name(icon_name)
            if not identity_type:
                self._gateway_label.set_text(identity_name)
            else:
                self._gateway_label.set_text(
                    f"{identity_name} ({identity_type.upper()})"
                )

        if result.supports(Namespace.REGISTER):
            self._register_button.set_sensitive(True)
            self._register_button.set_tooltip_text("")
        else:
            self._register_button.set_sensitive(False)
            self._register_button.set_tooltip_text(
                _("This gateway does not support direct registering.")
            )

        if result.supports(Namespace.COMMANDS):
            self._commands_button.set_sensitive(True)
            self._commands_button.set_tooltip_text("")
        else:
            self._commands_button.set_sensitive(False)
            self._commands_button.set_tooltip_text(
                _("This gateway does not support Ad-Hoc Commands.")
            )

    def _on_register_clicked(self, _button: Gtk.Button) -> None:
        assert self._result is not None
        open_window(
            "ServiceRegistration", account=self._account, address=self._result.jid
        )

    def _on_command_clicked(self, _button: Gtk.Button) -> None:
        assert self._result is not None
        open_window(
            "AdHocCommands", account=self._account, jids=[str(self._result.jid)]
        )


class AddContactGroupChatPage(AssistantPage):
    def __init__(self) -> None:
        AssistantPage.__init__(self)
        self.title = _("Join Group Chat?")

        self._result: DiscoInfo | None = None

        heading = Gtk.Label(
            label=_("Join Group Chat?"),
            wrap=True,
            max_width_chars=30,
            halign=Gtk.Align.CENTER,
            justify=Gtk.Justification.CENTER,
        )
        heading.add_css_class("title-1")
        self.append(heading)

        self._info_box = GroupChatInfoScrolled(minimal=True)
        self.append(self._info_box)

    def get_visible_buttons(self) -> list[str]:
        return ["back", "join"]

    def get_default_button(self) -> str:
        return "join"

    def prepare(self, account: str, result: DiscoInfo) -> None:
        self._result = result

        self._info_box.set_account(account)
        self._info_box.set_from_disco_info(result)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging
from io import BytesIO
from pathlib import Path

from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import GLib
from PIL import Image
from PIL import ImageFile
from PIL import UnidentifiedImageError

log = logging.getLogger("gajim.c.util.image")

ImageFile.LOAD_TRUNCATED_IMAGES = True


def get_texture_from_file(
    path: str | Path, size: int | None = None
) -> Gdk.Texture | None:
    pixbuf = get_pixbuf_from_file(path, size)
    if pixbuf is None:
        return None
    return Gdk.Texture.new_for_pixbuf(pixbuf)


def get_pixbuf_from_file(
    path: str | Path, size: int | None = None
) -> GdkPixbuf.Pixbuf | None:
    """Tries to load Pixbuf from file to return it with size.
    If Pixbuf fails, Pillow is used as fallback.
    """
    try:
        if size is None:
            return GdkPixbuf.Pixbuf.new_from_file(str(path))
        return GdkPixbuf.Pixbuf.new_from_file_at_scale(str(path), size, size, True)

    except GLib.Error:
        try:
            with open(path, "rb") as image_handle:
                img = Image.open(image_handle)
                converted_image = img.convert("RGBA")
        except (NameError, OSError, UnidentifiedImageError):
            log.warning("Pillow convert failed: %s", path)
            log.debug("Error", exc_info=True)
            return None

        array = GLib.Bytes.new(converted_image.tobytes())
        width, height = converted_image.size
        pixbuf = GdkPixbuf.Pixbuf.new_from_bytes(
            array, GdkPixbuf.Colorspace.RGB, True, 8, width, height, width * 4
        )
        if size is not None:
            width, height = scale_with_ratio(size, width, height)
            return pixbuf.scale_simple(width, height, GdkPixbuf.InterpType.BILINEAR)
        return pixbuf

    except RuntimeError as error:
        log.warning("Loading pixbuf failed: %s", error)
        return None


def get_texture_from_data(data: bytes, size: int | None = None) -> Gdk.Texture | None:
    """
    Load data into PixbufLoader and return GdkPixbuf.Pixbuf if possible.
    Pillow is used as fallback mechanism.
    """
    pixbufloader = GdkPixbuf.PixbufLoader()
    try:
        pixbufloader.write(data)
        pixbufloader.close()
        pixbuf = pixbufloader.get_pixbuf()
    except GLib.Error:
        pixbufloader.close()

        log.warning(
            "Loading image data using PixbufLoader failed. "
            "Trying to convert image data using Pillow."
        )
        try:
            image = Image.open(BytesIO(data)).convert("RGBA")
            array = GLib.Bytes.new(image.tobytes())
            width, height = image.size
            pixbuf = GdkPixbuf.Pixbuf.new_from_bytes(
                array, GdkPixbuf.Colorspace.RGB, True, 8, width, height, width * 4
            )
            image.close()
        except Exception:
            log.warning(
                "Could not use Pillow to convert image data. Image cannot be displayed",
                exc_info=True,
            )
            return None

    if pixbuf is None:
        return None

    pixbuf = pixbuf.apply_embedded_orientation()
    assert pixbuf is not None

    if size is not None:
        width, height = scale_with_ratio(size, pixbuf.get_width(), pixbuf.get_height())
        pixbuf = pixbuf.scale_simple(width, height, GdkPixbuf.InterpType.BILINEAR)
        assert pixbuf is not None

    return Gdk.Texture.new_for_pixbuf(pixbuf)


def scale_with_ratio(size: int, width: int, height: int) -> tuple[int, int]:
    if height == width:
        return size, size
    if height > width:
        ratio = height / float(width)
        return int(size / ratio), size

    ratio = width / float(height)
    return size, int(size / ratio)


def image_size(image_path: Path) -> tuple[int, int]:
    try:
        with Image.open(image_path) as image:
            width, height = image.size
            return width, height
    except Exception:
        log.warning("Could not determine image size for %s", image_path, exc_info=True)
        return 0, 0


def is_image_animated(image_path: Path) -> bool:
    try:
        with Image.open(image_path) as image:
            n_frames: int = getattr(image, "n_frames", 1)
            return getattr(image, "is_animated", False) and n_frames > 1
    except Exception:
        log.warning(
            "Could not determine if image is animated: %s", image_path, exc_info=True
        )
        return False



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Literal
from typing import overload
from typing import TYPE_CHECKING

from gajim.common import modules

if TYPE_CHECKING:
    from gajim.common.modules.adhoc_commands import AdHocCommands
    from gajim.common.modules.annotations import Annotations
    from gajim.common.modules.bits_of_binary import BitsOfBinary
    from gajim.common.modules.blocking import Blocking
    from gajim.common.modules.bookmarks import Bookmarks
    from gajim.common.modules.bytestream import Bytestream
    from gajim.common.modules.caps import Caps
    from gajim.common.modules.carbons import Carbons
    from gajim.common.modules.chat_markers import ChatMarkers
    from gajim.common.modules.chatstates import Chatstate
    from gajim.common.modules.contacts import Contacts
    from gajim.common.modules.discovery import Discovery
    from gajim.common.modules.entity_time import EntityTime
    from gajim.common.modules.gateway import Gateway
    from gajim.common.modules.http_auth import HTTPAuth
    from gajim.common.modules.http_muc_search import HttpMucSearch
    from gajim.common.modules.httpupload import HTTPUpload
    from gajim.common.modules.ibb import IBB
    from gajim.common.modules.iq import Iq
    from gajim.common.modules.jingle import Jingle
    from gajim.common.modules.last_activity import LastActivity
    from gajim.common.modules.mam import MAM
    from gajim.common.modules.mds import MDS
    from gajim.common.modules.message import Message
    from gajim.common.modules.moderations import Moderations
    from gajim.common.modules.muc import MUC
    from gajim.common.modules.muc_blocking import MucBlocking
    from gajim.common.modules.omemo import OMEMO
    from gajim.common.modules.pep import PEP
    from gajim.common.modules.ping import Ping
    from gajim.common.modules.presence import Presence
    from gajim.common.modules.pubsub import PubSub
    from gajim.common.modules.reactions import Reactions
    from gajim.common.modules.receipts import Receipts
    from gajim.common.modules.register import Register
    from gajim.common.modules.retraction import Retraction
    from gajim.common.modules.roster import Roster
    from gajim.common.modules.roster_item_exchange import RosterItemExchange
    from gajim.common.modules.search import Search
    from gajim.common.modules.security_labels import SecLabels
    from gajim.common.modules.software_version import SoftwareVersion
    from gajim.common.modules.user_avatar import UserAvatar
    from gajim.common.modules.user_location import UserLocation
    from gajim.common.modules.user_nickname import UserNickname
    from gajim.common.modules.user_tune import UserTune
    from gajim.common.modules.vcard4 import VCard4
    from gajim.common.modules.vcard_avatars import VCardAvatars
    from gajim.common.modules.vcard_temp import VCardTemp


class ClientModules:
    def __init__(self, account: str) -> None:
        self._account = account

    @overload
    def get_module(self, name: Literal['AdHocCommands']) -> AdHocCommands: ...
    @overload
    def get_module(self, name: Literal['Annotations']) -> Annotations: ...
    @overload
    def get_module(self, name: Literal['BitsOfBinary']) -> BitsOfBinary: ...
    @overload
    def get_module(self, name: Literal['Blocking']) -> Blocking: ...
    @overload
    def get_module(self, name: Literal['Bookmarks']) -> Bookmarks: ...
    @overload
    def get_module(self, name: Literal['Bytestream']) -> Bytestream: ...
    @overload
    def get_module(self, name: Literal['Caps']) -> Caps: ...
    @overload
    def get_module(self, name: Literal['Carbons']) -> Carbons: ...
    @overload
    def get_module(self, name: Literal['ChatMarkers']) -> ChatMarkers: ...
    @overload
    def get_module(self, name: Literal['Chatstate']) -> Chatstate: ...
    @overload
    def get_module(self, name: Literal['Contacts']) -> Contacts: ...
    @overload
    def get_module(self, name: Literal['Discovery']) -> Discovery: ...
    @overload
    def get_module(self, name: Literal['EntityTime']) -> EntityTime: ...
    @overload
    def get_module(self, name: Literal['Gateway']) -> Gateway: ...
    @overload
    def get_module(self, name: Literal['HTTPAuth']) -> HTTPAuth: ...
    @overload
    def get_module(self, name: Literal['HttpMucSearch']) -> HttpMucSearch: ...
    @overload
    def get_module(self, name: Literal['HTTPUpload']) -> HTTPUpload: ...
    @overload
    def get_module(self, name: Literal['IBB']) -> IBB: ...
    @overload
    def get_module(self, name: Literal['Iq']) -> Iq: ...
    @overload
    def get_module(self, name: Literal['Jingle']) -> Jingle: ...
    @overload
    def get_module(self, name: Literal['LastActivity']) -> LastActivity: ...
    @overload
    def get_module(self, name: Literal['MAM']) -> MAM: ...
    @overload
    def get_module(self, name: Literal['MDS']) -> MDS: ...
    @overload
    def get_module(self, name: Literal['Message']) -> Message: ...
    @overload
    def get_module(self, name: Literal['Moderations']) -> Moderations: ...
    @overload
    def get_module(self, name: Literal['MUC']) -> MUC: ...
    @overload
    def get_module(self, name: Literal['MucBlocking']) -> MucBlocking: ...
    @overload
    def get_module(self, name: Literal['OMEMO']) -> OMEMO: ...
    @overload
    def get_module(self, name: Literal['PEP']) -> PEP: ...
    @overload
    def get_module(self, name: Literal['Ping']) -> Ping: ...
    @overload
    def get_module(self, name: Literal['Presence']) -> Presence: ...
    @overload
    def get_module(self, name: Literal['PubSub']) -> PubSub: ...
    @overload
    def get_module(self, name: Literal['Reactions']) -> Reactions: ...
    @overload
    def get_module(self, name: Literal['Receipts']) -> Receipts: ...
    @overload
    def get_module(self, name: Literal['Register']) -> Register: ...
    @overload
    def get_module(self, name: Literal['Retraction']) -> Retraction: ...
    @overload
    def get_module(self, name: Literal['RosterItemExchange']) -> RosterItemExchange: ...  # noqa: E501
    @overload
    def get_module(self, name: Literal['Roster']) -> Roster: ...
    @overload
    def get_module(self, name: Literal['Search']) -> Search: ...
    @overload
    def get_module(self, name: Literal['SecLabels']) -> SecLabels: ...
    @overload
    def get_module(self, name: Literal['SoftwareVersion']) -> SoftwareVersion: ...  # noqa: E501
    @overload
    def get_module(self, name: Literal['UserAvatar']) -> UserAvatar: ...
    @overload
    def get_module(self, name: Literal['UserLocation']) -> UserLocation: ...
    @overload
    def get_module(self, name: Literal['UserNickname']) -> UserNickname: ...
    @overload
    def get_module(self, name: Literal['UserTune']) -> UserTune: ...
    @overload
    def get_module(self, name: Literal['VCardAvatars']) -> VCardAvatars: ...
    @overload
    def get_module(self, name: Literal['VCardTemp']) -> VCardTemp: ...
    @overload
    def get_module(self, name: Literal['VCard4']) -> VCard4: ...

    def get_module(self, name: modules.ModulesLiteralT) -> modules.ModulesT:
        return modules.get_module(self._account, name)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import re
import string
from dataclasses import dataclass
from dataclasses import field
from re import Match

from gi.repository import GLib

from gajim.common import regex
from gajim.common.util.jid import validate_jid
from gajim.common.util.text import escape_iri_query
from gajim.common.util.uri import InvalidUri
from gajim.common.util.uri import MailUri
from gajim.common.util.uri import parse_uri as analyze_uri
from gajim.common.util.uri import XmppIri

PRE = '`'
STRONG = '*'
STRIKE = '~'
EMPH = '_'

WHITESPACE = set(string.whitespace)
SPAN_DIRS = {PRE, STRONG, STRIKE, EMPH}
VALID_SPAN_START = WHITESPACE | SPAN_DIRS

PRE_RX = r'(?P<pre>^```.+?(^```$(.|\Z)))'
PRE_NESTED_RX = r'(?P<pre>^```.+?((^```$(.|\Z))|\Z))'
QUOTE_RX = r'(?P<quote>^(?=>).*?(^|\Z)(?!>))'

BLOCK_RX = re.compile(PRE_RX + '|' + QUOTE_RX, re.DOTALL | re.MULTILINE)
BLOCK_NESTED_RX = re.compile(PRE_NESTED_RX + '|' + QUOTE_RX, re.DOTALL | re.MULTILINE)
UNQUOTE_RX = re.compile(r'^> |^>', re.MULTILINE)

URI_OR_JID_RX = re.compile(
    fr'(?P<uri>(?<![\w+.-]){regex.IRI})|(?P<jid>{regex.XMPP.jid})')

SD = 0
SD_POS = 1
MAX_QUOTE_LEVEL = 20


@dataclass
class StyleObject:
    start: int
    end: int
    text: str


@dataclass
class BaseHyperlink(StyleObject):
    name: str
    uri: str
    start_byte: int
    end_byte: int

    def get_markup_string(self) -> str:
        href = GLib.markup_escape_text(self.uri)
        text = GLib.markup_escape_text(self.text)
        return f'<a href="{href}">{text}</a>'


@dataclass
class Hyperlink(BaseHyperlink):
    name: str = field(default='uri', init=False)


@dataclass
class Address(BaseHyperlink):
    name: str = field(default='address', init=False)


@dataclass
class XMPPAddress(BaseHyperlink):
    name: str = field(default='xmppadr', init=False)


@dataclass
class MailAddress(BaseHyperlink):
    name: str = field(default='mailadr', init=False)


@dataclass
class Block(StyleObject):

    @classmethod
    def from_match(cls: Any, match: Match[str]) -> Block:
        return cls(start=match.start(),
                   end=match.end(),
                   text=match.group(cls.name))


@dataclass
class PlainBlock(Block):
    name: str = field(default='plain', init=False)
    spans: list[Span] = field(default_factory=list['Span'])
    uris: list[BaseHyperlink] = field(default_factory=list['BaseHyperlink'])

    @classmethod
    def from_quote_match(cls, match: Match[str]) -> PlainBlock:
        return cls(start=match.start(),
                   end=match.end(),
                   text=match.group('quote'))


@dataclass
class PreBlock(Block):
    name: str = field(default='pre', init=False)


@dataclass
class QuoteBlock(Block):
    name: str = field(default='quote', init=False)
    blocks: list[Block] = field(default_factory=list['Block'])

    def unquote(self) -> str:
        return UNQUOTE_RX.sub('', self.text)


@dataclass
class Span(StyleObject):
    start_byte: int
    end_byte: int
    name: str


@dataclass
class PlainSpan(Span):
    name: str = field(default='plain', init=False)


@dataclass
class StrongSpan(Span):
    name: str = field(default='strong', init=False)


@dataclass
class EmphasisSpan(Span):
    name: str = field(default='emphasis', init=False)


@dataclass
class PreTextSpan(Span):
    name: str = field(default='pre', init=False)


@dataclass
class StrikeSpan(Span):
    name: str = field(default='strike', init=False)


SPAN_CLS_DICT = {
    STRONG: StrongSpan,
    EMPH: EmphasisSpan,
    PRE: PreTextSpan,
    STRIKE: StrikeSpan
}


@dataclass
class ParsingResult:
    text: str
    blocks: list[Block]


def find_byte_index(text: str, index: int):
    byte_index = -1
    for index_, c in enumerate(text):
        byte_index += len(c.encode())
        if index == index_:
            return byte_index

    raise ValueError(f'index not in string: {text}, {index}')


def process(text: str | bytes, level: int = 0) -> ParsingResult:
    if isinstance(text, bytes):
        text = text.decode()

    blocks = _parse_blocks(text, level)
    for block in blocks:
        if isinstance(block, PlainBlock):
            offset = 0
            offset_bytes = 0
            for line in block.text.splitlines(keepends=True):
                block.spans += _parse_line(line, offset, offset_bytes)
                block.uris += _parse_uris(line, offset, offset_bytes)

                offset += len(line)
                offset_bytes += len(line.encode())

        if isinstance(block, QuoteBlock):
            result = process(block.unquote(), level=level + 1)
            block.blocks = result.blocks

    return ParsingResult(text, blocks)


def process_uris(text: str | bytes) -> list[BaseHyperlink]:
    if isinstance(text, bytes):
        text = text.decode()

    uris: list[BaseHyperlink] = []
    offset = 0
    offset_bytes = 0
    for line in text.splitlines(keepends=True):
        uris += _parse_uris(line, offset, offset_bytes)
        offset += len(line)
        offset_bytes += len(line.encode())

    return uris


def _parse_blocks(text: str, level: int) -> list[Block]:
    blocks: list[Block] = []
    text_len = len(text)
    last_end_pos = 0

    rx = BLOCK_NESTED_RX if level > 0 else BLOCK_RX

    for match in rx.finditer(text):
        if match.start() != last_end_pos:
            blocks.append(PlainBlock(
                start=last_end_pos,
                end=match.start(),
                text=text[last_end_pos:match.start()]))

        last_end_pos = match.end()
        group_dict = match.groupdict()

        if group_dict.get('quote') is not None:
            if level > MAX_QUOTE_LEVEL:
                blocks.append(PlainBlock.from_quote_match(match))
            else:
                blocks.append(QuoteBlock.from_match(match))

        if group_dict.get('pre') is not None:
            blocks.append(PreBlock.from_match(match))

    if last_end_pos != text_len:
        blocks.append(PlainBlock(
            start=last_end_pos,
            end=text_len,
            text=text[last_end_pos:]))

    return blocks


def _parse_line(line: str, offset: int, offset_bytes: int) -> list[Span]:
    index: int = 0
    length = len(line)
    stack: list[tuple[str, int]] = []
    spans: list[Span] = []

    while index < length:
        sd = line[index]
        if sd not in SPAN_DIRS:
            index += 1
            continue

        is_valid_start = _is_valid_span_start(line, index)
        is_valid_end = _is_valid_span_end(line, index)

        if is_valid_start and is_valid_end:
            # Favor end over new start, this means parsing is done non-greedy
            if sd in [open_sd for open_sd, _ in stack]:
                is_valid_start = False

        if is_valid_start:
            if sd == PRE:
                index = _handle_pre_span(line,
                                         index,
                                         offset,
                                         offset_bytes,
                                         spans)
                continue

            stack.append((sd, index))
            index += 1
            continue

        if is_valid_end:
            if sd not in [open_sd for open_sd, _ in stack]:
                index += 1
                continue

            if _is_span_empty(sd, index, stack):
                stack.pop()
                index += 1
                continue

            start_pos = _find_span_start_position(sd, stack)
            spans.append(_make_span(line,
                                    sd,
                                    start_pos,
                                    index,
                                    offset,
                                    offset_bytes))

        index += 1

    return spans


def _parse_uris(line: str,
                offset: int,
                offset_bytes: int) -> list[BaseHyperlink]:
    uris: list[BaseHyperlink] = []

    def make(start: int, end: int, is_jid: bool) -> BaseHyperlink | None:
        if line[end - 1] == ',':
            # Trim one trailing comma
            end -= 1
        if '(' in line[:start] and line[end - 1] == ')':
            # Trim one trailing closing parenthesis if the match is preceded
            # by an opening one somewhere on the line
            end -= 1
        if re.fullmatch('[^:]+:', line[start:end]):
            # URIs that consist only of a scheme are thusly excluded
            return None
        return _make_hyperlink(line,
                               start,
                               end - 1,
                               offset,
                               offset_bytes,
                               is_jid)

    for match in URI_OR_JID_RX.finditer(line):
        start, end = match.span()
        hyperlink = make(start, end, bool(match.group('jid')))
        if hyperlink:
            uris.append(hyperlink)

    return uris


def _handle_pre_span(line: str,
                     index: int,
                     offset: int,
                     offset_bytes: int,
                     spans: list[Span]) -> int:

    # Scan ahead for the end
    end = line.find(PRE, index + 1)
    if end == -1:
        return index + 1

    if end - index == 1:
        # empty span
        return index + 1

    spans.append(_make_span(line, PRE, index, end, offset, offset_bytes))
    return end + 1


def _make_span(line: str,
               sd: str,
               start: int,
               end: int,
               offset: int,
               offset_bytes: int) -> Span:

    text = line[start:end + 1]

    start_byte = find_byte_index(line, start) + offset_bytes
    end_byte = find_byte_index(line, end) + offset_bytes + 1

    start += offset
    end += offset + 1
    span_class = SPAN_CLS_DICT.get(sd)
    assert span_class is not None

    return span_class(start=start,
                      start_byte=start_byte,
                      end=end,
                      end_byte=end_byte,
                      text=text)


def _make_hyperlink(line: str,
                    start: int,
                    end: int,
                    offset: int,
                    offset_bytes: int,
                    is_jid: bool) -> BaseHyperlink | None:

    text = line[start:end + 1]

    start_byte = find_byte_index(line, start) + offset_bytes
    end_byte = find_byte_index(line, end) + offset_bytes + 1

    start += offset
    end += offset + 1

    uri = text
    if is_jid:
        cls_ = Address
        try:
            validate_jid(text)
        except ValueError:
            return None
        uri = escape_iri_query(text)

    else:
        uri = text
        auri = analyze_uri(uri)
        match auri:
            case XmppIri():
                cls_ = XMPPAddress
            case MailUri():
                cls_ = MailAddress
            case InvalidUri():
                return None
            case _:
                cls_ = Hyperlink

    return cls_(start=start,
                start_byte=start_byte,
                end=end,
                end_byte=end_byte,
                uri=uri,
                text=text)


def _is_span_empty(sd: str, index: int, stack: list[tuple[str, int]]) -> bool:
    start_sd = stack[-1][SD]
    if start_sd != sd:
        return False

    pos = stack[-1][SD_POS]
    return pos == index - 1


def _find_span_start_position(sd: str, stack: list[tuple[str, int]]) -> int:
    while stack:
        start_sd, pos = stack.pop()
        if start_sd == sd:
            return pos

    raise ValueError('Unable to find opening span')


def _is_valid_span_start(line: str, index: int) -> bool:
    '''
    https://xmpp.org/extensions/xep-0393.html#span

    ... The opening styling directive MUST be located at the beginning
    of the line, after a whitespace character, or after a different opening
    styling directive. The opening styling directive MUST NOT be followed
    by a whitespace character ...
    '''

    try:
        char = line[index + 1]
    except IndexError:
        return False

    if char in WHITESPACE:
        return False

    if index == 0:
        return True

    char = line[index - 1]
    return char in VALID_SPAN_START


def _is_valid_span_end(line: str, index: int) -> bool:
    '''
    https://xmpp.org/extensions/xep-0393.html#span

    ... and the closing styling directive MUST NOT be preceded
    by a whitespace character ...
    '''

    if index == 0:
        return False

    char = line[index - 1]
    return char not in WHITESPACE



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import sys

import gajim


def get_extended_app_version() -> str:
    if gajim.IS_FLATPAK:
        package = "Flatpak"
    elif gajim.IS_FLATPAK_NIGHTLY:
        package = "Flatpak Nightly"
    elif gajim.IS_PORTABLE:
        package = "Windows Portable"
    elif gajim.IS_MS_STORE:
        package = "Windows Store"
    elif sys.platform == "win32":
        package = "Windows"
    elif sys.platform == "darwin":
        package = "macOS"
    else:
        package = "Unix/Linux"

    return f"{gajim.__version__} ({package})"



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# Roster

from __future__ import annotations

from typing import cast

from collections.abc import Iterable
from collections.abc import Iterator

import nbxmpp
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.structs import IqProperties
from nbxmpp.structs import RosterData
from nbxmpp.structs import RosterItem
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import Task

from gajim.common import app
from gajim.common import types
from gajim.common.events import RosterPush
from gajim.common.events import RosterReceived
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import BareContact


class Roster(BaseModule):

    _nbxmpp_extends = 'Roster'
    _nbxmpp_methods = [
        'delete_item',
        'set_item',
    ]

    def __init__(self, client: types.Client) -> None:
        BaseModule.__init__(self, client)

        self.handlers = [
            StanzaHandler(name='iq',
                          callback=self._process_roster_push,
                          typ='set',
                          ns=Namespace.ROSTER),
        ]

        self._roster: dict[JID, RosterItem] = {}

        # Groups cache for performance
        self._groups = None

    def load_roster(self) -> None:
        self._log.info('Load from database')
        roster = app.storage.cache.load_roster(self._account)
        if not roster:
            self._log.info('Database empty, reset roster version')
            app.settings.set_account_setting(
                self._account, 'roster_version', '')
            return

        for jid in roster:
            self._con.get_module('Contacts').add_contact(jid)

        self._roster = roster

    def _store_roster(self) -> None:
        app.storage.cache.store_roster(self._account, self._roster)

    def get_size(self) -> int:
        return len(self._roster)

    def request_roster(self) -> None:
        version = app.settings.get_account_setting(self._account,
                                                   'roster_version')

        self._log.info('Request version: %s', version)
        self._nbxmpp('Roster').request_roster(
            version, callback=self._on_request_roster)

    def _on_request_roster(self, task: Task) -> None:
        try:
            roster = cast(RosterData, task.finish())
        except Exception as error:
            self._log.warning(error)
            return

        self._log.info('Received Roster, version: %s', roster.version)

        if roster.version is None or roster.items is not None:
            # version is None:
            # ---------------
            # No Roster versioning supported this
            # means we got the complete roster
            #
            # items is not None:
            # ---------------
            # Roster versioning supported but
            # server opted to send us the whole roster
            assert roster.items is not None
            self._set_roster_from_data(roster.items)

        app.settings.set_account_setting(self._account,
                                         'roster_version',
                                         roster.version)

        app.ged.raise_event(RosterReceived(account=self._account))

        self._con.connect_machine()

    def _set_roster_from_data(self, items: list[RosterItem]) -> None:
        self._roster.clear()
        self._groups = None

        for item in items:
            self._log.info(item)
            self._con.get_module('Contacts').add_contact(item.jid)
            self._roster[item.jid] = item

        self._store_roster()

    def _process_roster_push(self,
                             _con: types.NBXMPPClient,
                             _stanza: Iq,
                             properties: IqProperties
                             ) -> None:
        self._log.info('Push received')
        assert properties.roster is not None
        item = properties.roster.item
        if item.subscription == 'remove':
            self._roster.pop(item.jid)
        else:
            self._roster[item.jid] = item

        self._groups = None
        self._store_roster()

        self._log.info('New version: %s', properties.roster.version)
        app.settings.set_account_setting(self._account,
                                         'roster_version',
                                         properties.roster.version)

        app.ged.raise_event(RosterPush(account=self._account,
                                       item=item))

        contact = self._con.get_module('Contacts').get_contact(item.jid)
        contact.notify('nickname-update')

        raise nbxmpp.NodeProcessed

    def get_item(self, jid: JID) -> RosterItem | None:
        return self._roster.get(jid)

    def set_groups(self, jid: JID, groups: Iterable[str] | None) -> None:
        if groups is not None:
            groups = set(groups)
        item = self.get_item(jid)
        assert item is not None
        self._nbxmpp('Roster').set_item(jid, item.name, groups)

    def get_groups(self) -> set[str]:
        if self._groups is not None:
            return set(self._groups)

        groups: set[str] = set()
        for item in self._roster.values():
            groups.update(item.groups)
        self._groups = groups
        return set(groups)

    def _get_items_with_group(self, group: str) -> list[RosterItem]:
        return list(filter(lambda item: group in item.groups,
                           self._roster.values()))

    def remove_group(self, group: str) -> None:
        items = self._get_items_with_group(group)
        for item in items:
            new_groups = item.groups - {group}
            self.set_groups(item.jid, new_groups)

    def rename_group(self, group: str, new_group: str) -> None:
        items = self._get_items_with_group(group)
        for item in items:
            new_groups = item.groups - {group}
            new_groups.add(new_group)
            self.set_groups(item.jid, new_groups)

    def change_group(self, jid: JID, old_group: str, new_group: str) -> None:
        item = self.get_item(jid)
        assert item is not None
        groups = set(item.groups)
        groups.discard(old_group)
        groups.add(new_group)
        self._nbxmpp('Roster').set_item(jid, item.name, groups)

    def add_to_group(self, jid: JID, group: str) -> None:
        item = self.get_item(jid)
        assert item is not None
        new_groups = item.groups | {group}
        self._nbxmpp('Roster').set_item(jid, item.name, new_groups)

    def remove_from_group(self, jid: JID, group: str) -> None:
        item = self.get_item(jid)
        assert item is not None
        new_groups = item.groups - {group}
        self._nbxmpp('Roster').set_item(jid, item.name, new_groups)

    def change_name(self, jid: JID, name: str | None) -> None:
        item = self.get_item(jid)
        assert item is not None
        self._nbxmpp('Roster').set_item(jid, name, item.groups)

    def iter(self) -> Iterator[tuple[JID, RosterItem]]:
        yield from self._roster.items()

    def iter_contacts(self) -> Iterator[BareContact]:
        for jid in self._roster:
            yield cast(BareContact, self._get_contact(jid))



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import functools
import logging
import queue
from collections.abc import Callable

from gi.repository import GLib

log = logging.getLogger('gajim.c.m.task_manager')


class TaskManager:
    def __init__(self) -> None:
        self._timeout: int | None = None
        self._queue: queue.PriorityQueue[Task] = queue.PriorityQueue()

    def _start_worker(self) -> None:
        self._timeout = GLib.timeout_add_seconds(2, self._process_queue)

    def _process_queue(self) -> bool:
        log.info('%s tasks queued', self._queue.qsize())
        requeue: list[Task] = []
        while not self._queue.empty():
            task = self._queue.get_nowait()
            if task.is_obsolete():
                log.info('Task obsolete: %r', task)
                continue

            if not task.preconditions_met():
                # precondition_met() can change the obsolete flag, so we need
                # to check again here
                if task.is_obsolete():
                    log.info('Task obsolete: %r', task)
                else:
                    requeue.append(task)
                continue

            log.info('Execute task %r', task)
            task.execute()
            self._requeue_tasks(requeue)
            return True

        if self._requeue_tasks(requeue):
            # Queue is empty, but there are tasks to requeue
            # don't stop worker
            return True

        # Queue is empty, stop worker
        self._timeout = None
        return False

    def _requeue_tasks(self, tasks: list[Task]) -> bool:
        if not tasks:
            return False

        for task in tasks:
            log.info('Requeue task (preconditions not met): %r', task)
            self._queue.put_nowait(task)
        return True

    def add_task(self, task: Task) -> None:
        log.info('Adding task: %r', task)
        self._queue.put_nowait(task)
        if self._timeout is None:
            self._start_worker()


@functools.total_ordering
class Task:  # noqa: PLW1641
    def __init__(self, priority: int = 0) -> None:
        self.priority = priority
        self._obsolete = False

    def is_obsolete(self) -> bool:
        return self._obsolete

    def set_obsolete(self) -> None:
        self._obsolete = True

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Task):
            raise NotImplementedError
        return self.priority < other.priority

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Task):
            raise NotImplementedError
        return other.priority == self.priority

    def execute(self) -> None:
        raise NotImplementedError

    def preconditions_met(self) -> bool:
        raise NotImplementedError


class PulseManager:
    def __init__(self) -> None:
        self._callbacks: list[Callable[[], Any]] = []
        self._timeout_id = GLib.timeout_add_seconds(60, self._execute_pulse)

    def add_callback(self, callback: Callable[[], Any]) -> None:
        self._callbacks.append(callback)

    def remove_callback(self, callback: Callable[[], Any]) -> None:
        self._callbacks.remove(callback)

    def _execute_pulse(self) -> bool:
        log.info('Execute pulse for %s callbacks', len(self._callbacks))
        for callback in self._callbacks:
            try:
                callback()
            except Exception:
                log.exception('Unable to execute pulse')

        return GLib.SOURCE_CONTINUE



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import Generic
from typing import TypeVar

from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Gtk


class BaseCompletionListItem:
    def get_text(self) -> str:
        raise NotImplementedError


L = TypeVar("L", bound=BaseCompletionListItem)


class BaseCompletionViewItem(Generic[L]):
    css_class = ""

    def __init__(self) -> None:
        self._bindings: list[GObject.Binding] = []

    def bind(self, obj: L) -> None:
        raise NotImplementedError

    def unbind(self) -> None:
        raise NotImplementedError

    def do_unroot(self) -> None:
        raise NotImplementedError


class BaseCompletionProvider:
    trigger_char = ""
    name = ""

    def get_model(self) -> tuple[Gio.ListModel[Any], type[BaseCompletionViewItem[Any]]]:
        raise NotImplementedError

    def check(self, candidate: str, start_iter: Gtk.TextIter) -> bool:
        raise NotImplementedError

    def populate(self, candidate: str, contact: Any) -> bool:
        raise NotImplementedError
